"""

    QuickAnalysis.Analyzer.py

    Copyright (c) 2016-2024, SAXS Team, KEK-PF

"""
import os
from molass_legacy.KekLib.BasicUtils import exe_name, clear_dirs_with_retry
import molass_legacy.KekLib.CustomMessageBox as MessageBox
from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker, log_exception
from molass_legacy._MOLASS.SerialSettings import get_setting, set_setting
from molass_legacy.SerialAnalyzer.DevSettings import get_dev_setting
from molass_legacy.QuickAnalysis.JudgeHolder import JudgeHolder

SD_DEBUG = False
if SD_DEBUG:
    from molass_legacy.SerialAnalyzer.SdDebugger import SdDebugger

class Analyzer:
    def __init__( self, main_dialog, loader ):
        self.dialog = main_dialog   # 
        self.parent = main_dialog   # temp fix
        self.loader = loader
        self.mapper_canvas  = None
        self.mapper_hook    = None
        self.app_logger     = None
        self.logger_visible = False
        self.cancel_reports = False     # True for testing only
        self.preview_survey = False
        self.waiting_for_reply = False  # used in test control
        self.mfc_dialog = None

    def set_cancel_mode(self):
        self.cancel_reports = True

    def set_preview_survey(self):
        self.preview_survey = True

    def has_mapper_canvas(self):
        return self.mapper_canvas is not None

    def set_mapper_hook( self, hook ):
        self.mapper_hook    = hook

    def change_log_to(self, analysis_folder):
        from molass_legacy.KekLib.ChangeableLogger import Logger

        self.app_logfile    = '%s/%s.log' % ( analysis_folder, exe_name() )
        if self.parent.tmp_logger is None:
            # TODO: investigate this case
            pass
        else:
            self.parent.tmp_logger.moveto( self.app_logfile )
            del self.parent.tmp_logger
            self.parent.tmp_logger = None
        self.app_logger     = Logger( self.app_logfile )

    def do_analysis(self, sd, pre_recog, md, analysis_folder, analysis_name, devel=False):
        # delayed imports to reduce initial load time
        from molass_legacy.KekLib.OurTkinter import Tk
        from molass_legacy.KekLib.ReadOnlyText import ReadOnlyText
        from molass_legacy.SerialAnalyzer.ProgressInfoUtil import estimate_init_max_dist
        from molass_legacy.KekLib.ProgressInfo import ProgressInfo
        from molass_legacy.KekLib.ProgressInfoDialog import ProgressInfoDialog
        # import Extrapolation
        from .AnalyzerDialogProxy import AnalyzerDialogProxy
        from molass_legacy.SerialAnalyzer.InputOutputDisplay import InputOutputDisplay
        from molass_legacy.SerialAnalyzer.AnalyzerUtil import compute_conc_factor_util
        from molass_legacy.Mapping.MapperConstructor import create_mapper        
        from molass_legacy.Mapping.ElutionMapperCanvas import ElutionMapperCanvas

        """
        in earlier versions than 20210604, pre_recog.set_info() has been repeated here.
        revised so as not to repeat that recognition
        """
        analysis_copy = sd._get_analysis_copy_impl(pre_recog)

        self.mapper_canvas  = None

        self.progress_dialog = None

        parent  = self.parent
        parent.config(cursor='wait')
        parent.update()

        self.data_folder    = get_setting( 'in_folder' ).replace( '\\', '/' )
        tester_info = self.dialog.tester_info
        if tester_info is None:
            tester_info_log = ''
        else:
            tester_info_log = ' with test pattern ' + str( tester_info.test_pattern )
        self.app_logger.info( "start analysis for " + self.data_folder + tester_info_log )

        self.set_controller_env( analysis_folder )

        self.analysis_copy = analysis_copy
        self.judge_holder = JudgeHolder(sd, pre_recog, analysis_copy)

        self.mapper = create_mapper(self.parent, self.analysis_copy, sd, pre_recog,
                                    callbacks=self.judge_holder.get_callbacks(),
                                    analyzer_dialog=self.parent, logger=self.app_logger)

        self.initial_navigation = True
        conc_factor = compute_conc_factor_util()

        sd_copy = self.analysis_copy.get_exec_copy(self.mapper, conc_factor)
        if SD_DEBUG:
            debugger = SdDebugger()
            debugger.save_info(self.analysis_copy)
            debugger.save_info(sd_copy)

        parent.config(cursor='')
        parent.update()

        self.mfc_dialog = None
        while True:
            if self.parent.use_mtd_conc:
                applied = self.do_microfluidic_analysis()
                if applied:
                    correction_necessity = False    # temporary
                else:
                    return
            else:
                self.mapper_canvas = ElutionMapperCanvas(self.dialog, sd_copy, sd, pre_recog, self.mapper, self.judge_holder, initial_navigation=self.initial_navigation )
                self.mapper_canvas.show()
                self.initial_navigation = False
                if self.mapper_canvas.applied:
                    # TODO
                    self.mapper =  self.mapper_canvas.mapper
                    correction_necessity = self.mapper_canvas.get_xray_correction_necessity()
                else:
                    return

            if SD_DEBUG:
                debugger.save_info(sd_copy)

            try:
                dialog = AnalyzerDialogProxy(parent, sd_copy, self.mapper, self.judge_holder, correction_necessity)
                dialog.apply()
            except:
                log_exception(self.app_logger, "AnalyzerDialogProxy Error: ")
                assert False

            if self.cancel_reports:
                self.cleanup_on_cancel()
                return

            break

        if self.mapper_canvas.ok_debugging:
            self.do_minimal_to_debug()
            return

        range_type      = 4
        min_value       = None
        preview_params = get_setting('preview_params')
        pdata, popts = preview_params
        if pdata.is_for_sec:
            # for compatibility with DecompEditorDialog concerning Preview
            analysis_range_info = get_setting('analysis_range_info')
            analysis_ranges = analysis_range_info.get_ranges()
        else:
            analysis_ranges = pdata.get_analysis_ranges()

        self.serial_data = sd_copy
        exec_sd = sd_copy
        mapped_info = self.mapper.get_mapped_info()

        parent.analysis_button.config( state=Tk.DISABLED )

        zx_flag = True

        scattering_correction = get_setting( 'scattering_correction' )
        num_files = len( exec_sd.datafiles )
        init_max_dict = estimate_init_max_dist( num_files, scattering_correction, analysis_ranges, zx_flag )

        pinfo = ProgressInfo( init_max_dict )
        stream_labels = [   "Baseline Correction  ",
                            "Guinier Analysis ",
                            "Extrapolation    " ]

        def description_cb( not_used_parent, frame ):
            dframe = Tk.Frame( frame )
            dframe.pack( fill=Tk.X, pady=10 )

            disp_label = Tk.Label( dframe, text="Input / Output", width=12, anchor=Tk.W )
            disp_label.grid( row=0, column=0, sticky=Tk.N + Tk.W )

            disp_frame = InputOutputDisplay( dframe, value_width=560 )
            disp_frame.grid( row=0, column=1, padx=5 )

        if self.parent.env_info.excel_is_available:
            text = None
            def plot_init_cb(not_used_parent, frame):
                nonlocal text
                text = ReadOnlyText(frame, relief=Tk.FLAT, height=12, bg='white', fg='orange')
                text.tag_config('CENTER_LARGE', justify=Tk.CENTER, font=('', 20))
                text.insert(Tk.END, "Caution:\n"
                                    "Do not open other Excel books by\n"
                                    "double-clicking during this execution\n",
                                    'CENTER_LARGE')
                text.tag_config('CENTER', justify=Tk.CENTER)
                text.insert(Tk.END, "\n"
                                    "since it may result in a confusion stated below.\n"
                                    "On double-click, Windows tries to use the Excel instance we created\n"
                                    "and it will be simultaneuosly controlled by the program and your manipulation\n"
                                    "and that can be confusing. It is hard to persuade Windows not to do that.",
                                    'CENTER')
                text.pack(fill=Tk.X)
            def plot_final_cb():
                text.pack_forget()
        else:
            plot_init_cb = None
            plot_final_cb = None

        self.boundary_test_deferred = False
        refresh_interval = 20 if get_dev_setting( 'use_simpleguinier' ) == 1 else 50
        self.progress_dialog = ProgressInfoDialog( self.dialog.parent,
                                    "Execution Progress", pinfo,
                                    stream_labels=stream_labels,
                                    description_cb=description_cb,
                                    plot_init_cb=plot_init_cb,
                                    refresh_log_cb=self.refresh_log_cb,
                                    plot_final_cb=plot_final_cb,
                                    is_alive_cb=self.is_alive_cb,
                                    refresh_interval=refresh_interval,
                                    logger=self.app_logger,
                                    )
        self.logger_visible = True

        def start_controller_closure():
            self.start_controller( exec_sd, range_type, analysis_ranges, preview_params, min_value, zx_flag, mapped_info )

        # start controller after progress_dialog.show to get all logs from controller
        parent.after( 100, start_controller_closure )

        btn_text = self.progress_dialog.show( logger=self.app_logger )
        self.logger_visible = False
        if btn_text == 'Cancel':
            self.controller.kill()
            self.controller.terminate()
            self.app_logger.info( 'Canceled.' )

        cleaner = self.parent.cleaner
        if cleaner is not None:
            cleaner.set_cleanup(None)
        self.progress_dialog = None
        self.app_logger = None      # release the log file at this timing
        parent.analysis_button.config( state=Tk.NORMAL )
        parent.update_plot_button_state()

    def get_final_std_diff( self ):
        # self.mapper_canvas.final_std_diff should have been set by show_mapping_figure_func call
        print( 'Analyzer: get_final_std_diff' )
        return self.mapper_canvas.final_std_diff

    def set_controller_env( self, analysis_folder ):
        conc_folder = get_setting( 'uv_folder' ).replace( '\\', '/' )
        self.conc_file   = get_setting( 'uv_file'   )
        if self.conc_file.find( '*' ) >= 0:
            self.conc_file = None
        outp_folder = analysis_folder
        self.conc_folder    = conc_folder
        self.work_folder    = outp_folder
        self.temp_folder    = self.work_folder + '/.temp'
        self.make_temp_folder()
        set_setting( 'temp_folder', self.temp_folder )
        self.guinier_folder = self.work_folder + '/.guinier_result'
        self.stamp_file     = os.path.join( self.guinier_folder, '--stamp.csv' )

    def make_temp_folder( self ):
        """
        TODO:
            unify with SerialController
        """
        try:
            clear_dirs_with_retry( [self.temp_folder] )
        except Exception as exc:
            etb = ExceptionTracebacker()
            self.app_logger.error( etb )
            raise exc

    def start_controller( self, serial_data, range_type, analysis_ranges, preview_params, min_value, zx_flag, mapped_info ):
        # delayed imports to reduce initial load time
        from molass_legacy.SerialAnalyzer.SerialControlInfo import SerialControlInfo
        from molass_legacy.SerialAnalyzer.SerialController import SerialController

        """
        temporarily commented out due to the following error
                pickle.dump( self.dictionary, pf )
            TypeError: cannot pickle '_tkinter.tkapp' object
        """
        # save_settings()     # make sure that settings be consistent in the execution thread

        result_book = get_setting( 'result_book' )
        book_file   = '/'.join( [ self.work_folder, result_book ] )
        serial_file = self.work_folder + '/.temp/--serial_result.csv'
            # TODO: unify .temp

        self.app_logger.info("ATSAS avalability is %s", str(self.parent.env_info.atsas_is_available))

        known_info_list = get_setting('known_info_list')
        control_info = SerialControlInfo(  self.data_folder,
                                        self.conc_folder,
                                        self.work_folder,
                                        self.temp_folder,
                                        conc_file=self.conc_file,
                                        serial_data=serial_data,
                                        guinier_folder=self.guinier_folder,
                                        stamp_file=self.stamp_file,
                                        serial_file=serial_file, book_file=book_file,
                                        zx=zx_flag,
                                        mapped_info=mapped_info,
                                        min_analysis_value=min_value,
                                        range_type=range_type,
                                        analysis_ranges=analysis_ranges,
                                        preview_params=preview_params,
                                        maintenance_log=self.parent.maintenance_log,
                                        env_info=self.parent.env_info,
                                        known_info_list=known_info_list,
                                        cleaner=self.parent.cleaner,
                                        )

        self.controller = SerialController(control_info)
        self.controller.start()

    def is_alive_cb(self):
        return self.controller.exe_thread.is_alive()

    def cleanup_on_cancel( self ):
        self.parent.update()
        self.app_logger = None      # release the log file when canceled
        log_size = os.path.getsize( self.app_logfile )
        yn = MessageBox.askyesno( "Log file removal confirmation",
                "Would you like to remove the log in %s (%d bytes)?" % ( self.app_logfile, log_size ),
                parent=self.parent )
        if yn:
            os.remove( self.app_logfile )

    def refresh_log_cb( self, text, level ):
        if text.find( 'deferred' ) >= 0:
            self.boundary_test_deferred = True

    def get_boundary_test_deferred( self ):
        return self.boundary_test_deferred

    def get_logfile( self ):
        return self.app_logfile

    def do_microfluidic_analysis(self):
        from Microfluidics.MctDecompDialog import MctDecompDialog
        xdata = self.parent.measured_data.get_xdata_for_mct()
        in_folder = self.data_folder
        self.mfc_dialog = MctDecompDialog(self.parent, xdata, in_folder)
        self.mfc_dialog.decompose()
        self.mfc_dialog.show()
        applied = self.mfc_dialog.applied
        self.mfc_dialog.destroy()
        return applied

    def do_minimal_to_debug(self):
        print("do_minimal_to_debug")



"""
    Proportional.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize
from scipy.interpolate import UnivariateSpline
from molass_legacy.KekLib.SciPyCookbook import smooth
from molass_legacy.Models.ElutionCurveModels import egha, emga

VARY_SMALL_VALUE = 1e-6

def get_curve_list(x, y, func, params):

    cy_list = []
    for p in params:
        cy = func(x, *p)
        cy_list.append(cy)

    ty = np.sum(cy_list, axis=0)
    ty[ty < VARY_SMALL_VALUE] = VARY_SMALL_VALUE

    ret_list = []
    for cy in cy_list:
        ret_list.append(y*cy/ty)

    return ret_list

def get_proportional_curves(x, y, egha_params, emga_params):
    egha_list = get_curve_list(x, y, egha, egha_params)
    emga_list = get_curve_list(x, y, emga, emga_params)
    return egha_list, emga_list



# coding: utf-8
"""
    SmoothnessPenaltyNumba.py

    Copyright (c) 2018-2020, SAXS Team, KEK-PF
"""

import numpy as np
try:
    # for numba 1.49 or later
    from numba.core.decorators import jit
except:
    from numba.decorators import jit

@jit(nopython=True)
def get_penalty_impl_numba(conc_dependence, P, penalty_weights, col_size):
    penalties = np.zeros(col_size)
    for j in range(col_size):
        pwi = j%conc_dependence
        if pwi > 1:
            continue

        x = P[:,j]
        penalty = 0
        for k in range(len(x)-1):
            penalty += (x[k] - x[k+1])**2
        penalties[j] = penalty * penalty_weights[pwi]
    return penalties

@jit(nopython=True)
def get_penalty_diff_impl_numba(conc_dependence, P, penalty_weights, col_size):
    NDP = np.zeros(P.shape)
    for j in range(col_size):
        pwi = j%conc_dependence
        if pwi > 1:
            continue

        scale = penalty_weights[pwi]
        x = P[:,j]
        NDP[ 0,j]   = scale*2*(x[0] - x[1])
        for k in range(1, len(x)-1):
            NDP[k,j]    = scale*2*(2*x[k] - x[k-1] - x[k+1])
        NDP[-1,j]   = -scale*2*(x[-2] - x[-1])

    return NDP

@jit(nopython=True)
def get_penalty_bq_ignore_impl_numba(P, penalty_weight, col_size):
    penalties = np.zeros(col_size)
    for j in range(col_size):
        x = P[:,j]
        penalty = 0
        for k in range(len(x)):
            penalty += (x[k] - x[k+1])**2
        penalties[j] = penalty * penalty_weight
    return penalties

@jit(nopython=True)
def get_penalty_diff_bq_ignore_impl_numba(P, weight, col_size):
    NDP = np.zeros(P.shape)
    for j in range(col_size):
        x = P[:,j]
        NDP[ 0,j]   = weight*2*(x[0] - x[1])
        for k in range(1, len(x)-1):
            NDP[k,j]    = weight*2*(2*x[k] - x[k-1] - x[k+1])
        NDP[-1,j]   = -weight*2*(x[-2] - x[-1])

    return NDP



"""

    ReducedCurve.py

    Copyright (c) 2023, SAXS Team, KEK-PF

"""
import numpy as np
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.SerialAnalyzer.ElutionCurve import ElutionCurve
from molass_legacy.Elution.CurveUtils import simple_plot

RECUDE_LIMIT_RATIO = 0.1

def make_reduced_curve(curve, debug=False):
    """
    temporary coping with the 20201007_2 trouble
    """
    red_curve = ElutionCurve(curve.y, x=curve.x)

    if debug:
        with plt.Dp():
            fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
            fig.suptitle("make_reduced_curve entry")
            simple_plot(ax1, curve)
            simple_plot(ax2, red_curve)
            fig.tight_layout()
            plt.show()

    max_y = curve.max_y
    red_peak_top_x = []
    red_peak_info = []
    red_boundaries = []
    primary_index = None
    max_py = None
    i = 0
    for k, info in enumerate(curve.peak_info):
        px = info[1]
        py = curve.spline(px)
        ratio = py/max_y
        print([k], "ratio=", ratio)
        if ratio < RECUDE_LIMIT_RATIO or px/curve.x[-1] > 0.9:
            # px/curve.x[-1] > 0.9 for ignoring the last peak of uv_curve in 20201007_2
            continue

        if max_py is None or py > max_py:
            max_py = py
            primary_index = i

        i += 1
        red_peak_top_x.append(curve.peak_top_x[k])
        red_peak_info.append(info)
        if k < len(curve.boundaries):
            red_boundaries.append(curve.boundaries[k])

    red_curve.primary_peak_no = primary_index
    red_curve.peak_top_x = np.array(red_peak_top_x)
    red_curve.peak_info = red_peak_info
    red_curve.boundaries = red_boundaries

    if debug:
        with plt.Dp():
            fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
            fig.suptitle("make_reduced_curve return")
            simple_plot(ax1, curve)
            simple_plot(ax2, red_curve)
            fig.tight_layout()
            plt.show()

    return red_curve



"""
    PeakRegion.py

    Copyright (c) 2021-2024, SAXS Team, KEK-PF
"""
import logging
import numpy as np
from bisect import bisect_right
from molass_legacy.KekLib.SciPyCookbook import smooth
from molass_legacy.DataStructure.LPM import get_corrected
from molass_legacy.Mapping.CurveSimilarity import CurveSimilarity

USE_ROTATION = True
if USE_ROTATION:
    from molass_legacy.KekLib.GeometryUtils import rotated_argminmax
    NORMAL_ANGLE = -np.pi/8
else:
    TAIL_RATIO = 0.15               # > 0.12 for Factin
    STRICT_RATIO = 0.004            # < 0.01 for 20190309_1, <= 0.004 for Sugiyama
IN_PEAK_DISMISS_RATIO = 0.3     # value to dismiss pH6, not to dismiss 20170301
SYMMETRY_ALLOW = 0.2            # > 0.12 for 20190529_1
SIGNIFICANT_RATIO_LIMIT = 0.05  # > 0.057 for 20161202, < 0.116 for SUB_TRN1, < 0.051 for Sugiyama
SPECIAL_CIRCUMVENT_LIMIT = 0.02 # > 0.0185 for 20170301
PTX_RATIO_LIMIT = 0.8
CORRECTION_THRESHOLD_RATIO = 0.3

class PeakRegion:
    def __init__(self, x_curve, a_curve, a_curve2, debug=False):
        self.logger = logging.getLogger(__name__)

        y = x_curve.y
        ratio = abs(y[0] - y[-1])/x_curve.max_y
        if ratio > CORRECTION_THRESHOLD_RATIO:
            # as in 20181203
            from molass_legacy.DataStructure.LPM import get_corrected
            from molass_legacy.SerialAnalyzer.ElutionCurve import ElutionCurve
            x = x_curve.x
            y = get_corrected(y, x=x)
            x_curve = ElutionCurve(y, x=x)
            self.logger.info("x_curve has been temporarily replaced by a corrected curve due to the ratio %.3g", ratio)

        self.x_curve = x_curve
        self.a_curve = a_curve
        self.a_curve2 = a_curve2

        self.similarity = CurveSimilarity(a_curve, x_curve)
        self.compute_cumsum(x_curve, a_curve)
        self.set_default_ends(debug=debug)
        self.set_wider_ends()

    def compute_cumsum(self, x_curve, a_curve):
        xr_cy, xr_y = self.compute_cumsum_impl(x_curve)
        ptx = x_curve.get_primarypeak_i()
        ptx_ratio = xr_cy[ptx]/xr_cy[-1]
        print("ptx_ratio=", ptx_ratio)
        if ptx_ratio > PTX_RATIO_LIMIT:
            # as in 20200304_1
            self.logger.info("re-computing cumsum for peak region due to ptx_ratio=%.3g > %.3g", ptx_ratio, PTX_RATIO_LIMIT)
            xr_cy = self.compute_cumsum_from_uv(x_curve, a_curve)
        self.xr_y = xr_y
        self.xr_cy = xr_cy

    def compute_cumsum_impl(self, curve, debug=False):
        xr_y = get_corrected(smooth(curve.y))    # smoothing is required in low quality data as in pH7
        xr_cy = np.cumsum(xr_y)
        if debug:
            import molass_legacy.KekLib.DebugPlot as plt
            with plt.Dp():
                x = curve.x
                y = curve.y
                fig, ax = plt.subplots()
                ax.set_title("compute_cumsum_impl")
                ax.plot(x, y)
                ax.plot(x, xr_y)
                axt = ax.twinx()
                axt.plot(x, xr_cy)
                fig.tight_layout()
                plt.show()
        return xr_cy, xr_y

    def compute_cumsum_from_uv(self, x_curve, a_curve, debug=False):
        from scipy.interpolate import UnivariateSpline

        uv_cy, uv_y = self.compute_cumsum_impl(a_curve)
        scale = x_curve.max_y/a_curve.max_y
        uv_cy_spline = UnivariateSpline(a_curve.x, scale * uv_cy, s=0, ext=3)
        uv_x = x_curve.x * self.similarity.slope + self.similarity.intercept
        xr_cy = uv_cy_spline(uv_x)
        if debug:
            import molass_legacy.KekLib.DebugPlot as plt
            from molass_legacy.Elution.CurveUtils import simple_plot
            with plt.Dp():
                fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
                simple_plot(ax1, a_curve, color="blue")
                simple_plot(ax2, x_curve, color="orange")
                ax1t = ax1.twinx()
                ax1t.grid(False)
                ax1t.plot(a_curve.x, uv_cy)
                ax2t = ax2.twinx()
                ax2t.grid(False)
                ax2t.plot(x_curve.x, xr_cy)
                fig.tight_layout()
                plt.show()
        return xr_cy

    def set_default_ends(self, debug=False, fig_file=None):
        if USE_ROTATION:
            xr_peak_ends, uv_peak_ends = self.get_peak_ends_with_rotation(NORMAL_ANGLE)
        else:
            xr_peak_ends, uv_peak_ends = self.get_peak_ends([TAIL_RATIO, 1 - TAIL_RATIO])
        self.set_peak_ends_with_traditional_infos(xr_peak_ends, uv_peak_ends, debug, fig_file)

    def set_wider_ends(self):
        if USE_ROTATION:
            pe = self.peak_ends
            half_width = (pe[1] - pe[0])*0.7
            center = (pe[1] + pe[0])/2
            uv_peak_ends = [ max(0, int(center - half_width)), min(len(self.a_curve.x)-1, int(center + half_width))]
        else:
            xr_peak_ends, uv_peak_ends = self.get_peak_ends([STRICT_RATIO, 1 - STRICT_RATIO])
        self.wider_ends = uv_peak_ends

    def get_peak_ends_impl(self, cy, proportions):
        peak_ends = []
        for r in proportions:
            j = bisect_right(cy, cy[-1]*r)
            peak_ends.append(j)
        return peak_ends

    def get_peak_ends(self, proportions, recursing=False):
        xr_cy = self.xr_cy
        xr_peak_ends = self.get_peak_ends_impl(xr_cy, proportions)
        xr_y = self.xr_y
        x_curve = self.x_curve

        if not USE_ROTATION:
            if xr_y[-1]/x_curve.max_y > TAIL_RATIO:
                # as in pH7
                xr_peak_ends[1] = len(xr_cy) - 1
                pass

        uv_peak_ends = self._guess_mapping(x_curve, xr_peak_ends, self.a_curve)

        if not recursing and not self.peak_ends_safety_check(self.a_curve, uv_peak_ends):
            self.logger.info("guessing peak region from molass_legacy.UV elution due to an unfavorable state.")
            # as in 20191118_3
            self.xr_cy = self.compute_cumsum_from_uv(x_curve, self.a_curve)
            xr_peak_ends, uv_peak_ends = self.get_peak_ends(proportions, recursing=True)

        return xr_peak_ends, uv_peak_ends

    def get_peak_ends_with_rotation(self, angle):
        xr_peak_ends = list(rotated_argminmax(angle, self.xr_cy, debug=False))
        uv_peak_ends = self._guess_mapping(self.x_curve, xr_peak_ends, self.a_curve)
        return xr_peak_ends, uv_peak_ends

    def _guess_mapping(self, x_curve, xr_peak_ends, a_curve):
        uv_peak_ends = [self.similarity.mapped_int_value(j) for j in xr_peak_ends]
        return uv_peak_ends

    def peak_ends_safety_check(self, ecurve, peak_ends, debug=False):
        # this is a risky judgement, any safer method is desired

        allow = 0       # setting safely for one can be harmful to the other as between 20171226 and 20191118_3
        ret = (peak_ends[0] - allow < ecurve.peak_info[0][1] and ecurve.peak_info[-1][1] < peak_ends[1] + allow)
                        # True for test_20191006_proteins5 and 20191118_3
        if debug:
            import molass_legacy.KekLib.DebugPlot as plt
            from molass_legacy.Elution.CurveUtils import simple_plot
            with plt.Dp():
                fig, ax = plt.subplots()
                simple_plot(ax, ecurve)
                ymin, ymax = ax.set_ylim()
                ax.set_ylim(ymin, ymax)
                for k, x in enumerate([peak_ends[0] - allow, ecurve.peak_info[0][1], ecurve.peak_info[-1][1], peak_ends[1] + allow]):
                    color = "yellow" if k in [0, 3] else "cyan"
                    ax.plot([x, x], [ymin, ymax], color=color)
                fig.tight_layout()
                plt.show()
        return ret

    def set_peak_ends_with_traditional_infos(self, xr_peak_ends, uv_peak_ends, debug, fig_file):
        x_curve = self.x_curve
        a_curve = self.a_curve
        a_curve2 = self.a_curve2

        first_peak, last_peak, uv_peak_ends_old, y = self._get_traditional_peak_ends(a_curve, curve2=a_curve2)
        xr_peak_ends_old = self._get_traditional_peak_ends(x_curve)[2]

        if debug:
            import molass_legacy.KekLib.DebugPlot as plt
            from molass_legacy.Elution.CurveUtils import simple_plot
            from molass_legacy.SerialAnalyzer.DataUtils import get_in_folder

            if USE_ROTATION:
                lables = ["rotated min", "rotated max"]
            else:
                lables = ["%.3g" % p for p in [TAIL_RATIO, 1 - TAIL_RATIO]]

            xr_y = self.xr_y
            xr_cy = self.xr_cy
            in_folder = get_in_folder()
            plt.push()
            fig, (ax0, ax1, ax2) = plt.subplots(ncols=3, figsize=(18,5))
            fig.suptitle("Peak Region Determination from Integral Proportions for %s" % in_folder, fontsize=20)
            ax0.set_title("UV Peak Region", fontsize=16)
            ax1.set_title("Xray Peak Region", fontsize=16)
            ax2.set_title("LPM-corrected Xray Elution and Integral Proportions", fontsize=16)

            uv_slice = slice(*uv_peak_ends)
            xr_slice = slice(*xr_peak_ends)
            for ax, curve, color, slice_ in [(ax0, a_curve, "blue", uv_slice), (ax1, x_curve, "orange", xr_slice)]:
                simple_plot(ax, curve, color=color)
                x_ = curve.x
                y_ = curve.y
                rx = x_[slice_]
                zy = np.zeros(len(rx))
                ry = y_[slice_]
                ax.fill_between(rx, zy, ry, fc='pink', alpha=0.2)

            ax2.plot(x_curve.x, xr_y, ":", color="orange")
            axt = ax2.twinx()
            axt.grid(False)
            axt.plot(x_curve.x, xr_cy)

            ymin0, ymax0 = ax0.get_ylim()
            ax0.set_ylim(ymin0, ymax0)
            ymin1, ymax1 = ax1.get_ylim()
            ax1.set_ylim(ymin1, ymax1)
            ymin2, ymax2 = ax2.get_ylim()
            ax2.set_ylim(ymin2, ymax2)

            for i, p in enumerate(uv_peak_ends_old + uv_peak_ends):
                color = "yellow" if i < 2 else "cyan"
                ax0.plot([p, p], [ymin0, ymax0], color=color)

            for i, p in enumerate(xr_peak_ends_old + xr_peak_ends):
                color = "yellow" if i < 2 else "cyan"
                ax1.plot([p, p], [ymin1, ymax1], color=color)
                ax2.plot([p, p], [ymin2, ymax2], color=color)

            xmint, xmaxt = axt.get_xlim()
            ymint, ymaxt = axt.get_ylim()
            for i, p in enumerate(xr_peak_ends):
                px = p
                py = xr_cy[p]
                dx = (xmaxt - xmint)*0.1*(-1 + 2*i)
                dy = (ymaxt - ymint)*0.1*(1 - 2*i)
                axt.annotate(lables[i], xy=(px, py), xytext=(px+dx, py+dy), ha='center', arrowprops=dict(arrowstyle="->", color='k'))

            fig.tight_layout()
            if fig_file is None:
                plt.show()
            else:
                from time import sleep
                plt.show(block=False)
                fig = plt.gcf()
                fig.savefig(fig_file)
                sleep(1)
            plt.pop()

        self.y = y
        self.first_peak = first_peak
        self.last_peak = last_peak

        excess_width_ratio = (xr_peak_ends_old[0] - xr_peak_ends_old[0])/(xr_peak_ends[1] - xr_peak_ends_old[1])
        symmetry_deviation = abs(excess_width_ratio - 1)
        print("excess_width_ratio=%.3g" % excess_width_ratio)
        print("symmetry_deviation=%.3g" % symmetry_deviation)
        if symmetry_deviation < SYMMETRY_ALLOW:
            # trick to discard anomaly in 20190529_1
            self.peak_ends = uv_peak_ends
            self.logger.info("peak_ends have been set from integral proportions due to symmetry_deviation=%.3g < %.3g", symmetry_deviation, SYMMETRY_ALLOW)
        else:
            self.peak_ends = uv_peak_ends

    def _get_traditional_peak_ends(self, curve, curve2=None):
        first_peak = curve.peak_info[0]
        if curve2 is None:
            y = None
        else:
            x = curve2.x
            y = curve2.y
            pos_ratio = first_peak[1]/len(x)
            # print("pos_ratio=", pos_ratio)
            if pos_ratio < 0.05:    # pos_ratio = 0.023 for Kosugi8, 0.127 for 20160227
                # as in Kosugi8
                j = first_peak[2]
                y_ = y.copy()
                y_[0:j] = y[j]
                y = y_
                first_peak = curve.peak_info[1]
                self.logger.info("the first peak with pos_ratio=%.3g has been replaced with the second peak.", pos_ratio)

        w = IN_PEAK_DISMISS_RATIO
        end_L = first_peak[0]*(1-w) + first_peak[1]*w
        last_peak = curve.peak_info[-1]
        end_R = last_peak[2]
        peak_ends = [end_L, end_R]
        return first_peak, last_peak, peak_ends, y

    def get_first_peak(self):
        return self.first_peak, self.y

    def get_ends(self):
        return self.peak_ends

    def get_size(self):
        return self.peak_ends[1] - self.peak_ends[0]

    def get_wider_ends(self):
        return self.wider_ends

    def get_slice(self):
        return slice(self.peak_ends[0], self.peak_ends[1]+1)

    def get_outside(self):
        ret = np.ones(len(self.y), dtype=bool)
        ret[self.get_slice()] = False
        return ret

    def is_in_the_region(self, x):
        return self.peak_ends[0] <= x and x <= self.peak_ends[1]

    def is_special(self, x, return_ratio=False):
        sig_ratio = self.get_significance_ratio(x)
        print("sig_ratio=", sig_ratio)
        ret_bool = self.is_in_the_wider_region(x) and sig_ratio > SIGNIFICANT_RATIO_LIMIT
        if ret_bool:
            ret, diff_ratio = self.can_be_treated_as_usual(x)
            if ret:
                self.logger.info("this case is circumvented as not special due to diff_ratio=%.3g", diff_ratio)
                ret_bool = False

        if return_ratio:
            return ret_bool, sig_ratio
        else:
            return ret_bool

    def is_in_the_wider_region(self, x):
        return self.wider_ends[0] <= x and x <= self.wider_ends[1]

    def get_significance_ratio(self, x):
        return self.a_curve.spline(x)/self.a_curve.max_y

    def get_distance_ratio(self, x):
        return (max(0, self.peak_ends[0] - x) + max(0, x - self.peak_ends[1]))/(self.peak_ends[1] - self.peak_ends[0])

    def can_be_treated_as_usual(self, x):
        if x < self.a_curve.primary_peak_x:
            y1 = self.a_curve.spline(x)
            y2 = self.a_curve.y[-1]
            diff_ratio = abs(y2 - y1)/self.a_curve.max_y
            ret_bool = diff_ratio < SPECIAL_CIRCUMVENT_LIMIT
        else:
            diff_ratio = None
            ret_bool = False
        return ret_bool, diff_ratio



"""

    PeakMapper.py

    Copyright (c) 2018-2024, SAXS Team, KEK-PF

"""
import copy
from bisect import bisect_right
import numpy            as np
import logging
from scipy.stats        import pearsonr, linregress
from scipy.interpolate  import UnivariateSpline
from scipy.optimize     import minimize
from itertools          import combinations
import OurStatsModels   as sm
from importlib import reload
from .SimplestMapper import SimplestMapper
from .PeakCurve import PeakCurve, get_peak_curve_info
from molass_legacy.SerialAnalyzer.ElutionCurve       import ElutionCurve
from molass_legacy.Elution.CurveUtils  import simple_plot, find_rotated_extreme_arg
from molass_legacy.KekLib.SciPyCookbook      import smooth
from molass_legacy.KekLib.BasicUtils         import Struct
from molass_legacy.KekLib.ExceptionTracebacker import log_exception, warnlog_exception
import molass_legacy.KekLib.DebugPlot        as plt
from molass_legacy._MOLASS.SerialSettings     import get_setting

COMPARE_ADDING_PEAKS_2WAYS  = True
NUM_PEAKS_SCORING_WEIGHT    = 0.01
USE_SCORED_BOUNDARY         = False
ROUGH_MIN_Y_RATIO_LIST      = [0.02, 0.1]
ADEQUACY_CORRELATION        = 0.997     # 0.996 for monoE, 0.99997 for SUB_TRN1
RISKY_CORRELATION_LIMIT     = 0.95      # > 0.88 for 20190221_2
NUM_ROOTS_TO_AVOID_NOISE    = 5         # to void low quality peaks as in pH7, > 3 for SUB_TRN1
FEASIBLE_DIFF_NUM           = 2
REDUCE_WHEN_INFEASIBLE      = True
ADOPTABLE_SCORE_RATIO       = 1.05      # > 1.01 for 20170304
RELIABLE_PEAK_SCORE         = 0.9
NUM_FRAMES_BOUNADRY         = 400       # < 461 for 20210628_2
MINIMUM_HEIGHT_RATIO        = 0.04      # > 0.03 for 20200121_2
ACCEPTABLE_SIMILARITY       = 0.97      # > 0.93 for 20210323_1, 0.96 for 20200121_2
DEEM_SINGLE_PEAK_RATIO      = 0.1
ACCEPTABLE_ENDS_RATIO_DEV   = 0.5

def debug_plot( a_curve, x_curve, where_str, params=None):
    from DataUtils import get_in_folder

    plt.push()
    fig = plt.figure(figsize=(14, 7))
    fig.suptitle( 'PeakMapper debug plot for ' + get_in_folder() + ' (%s)' % where_str, fontsize=24 )
    ax1 = fig.add_subplot( 121 )
    ax2 = fig.add_subplot( 122 )
    ax1.set_title( "UV Curve", fontsize=20 )
    ax2.set_title( "Xray Curve", fontsize=20 )

    ax1.plot( a_curve.y )
    ax1.plot( a_curve.spline(a_curve.x), color='lime' )
    ax2.plot( x_curve.y, color='orange' )
    ax2.plot( x_curve.spline(x_curve.x), color='lime' )

    def plot_peaks( ax, curve ):
        for info in curve.peak_info:
            peak = info[1]
            ax.plot( peak, curve.spline( peak ), 'o', color='red', markersize=10 )

        for b in curve.boundaries:
            ax.plot( b, curve.spline( b ), 'o', color='yellow', markersize=8  )

    plot_peaks( ax1, a_curve )
    plot_peaks( ax2, x_curve )

    if params is not None:
        A, B = params
        A_, B_ = 1/A, -B/A
        for x1 in a_curve.peak_top_x:
            x = A_*x1 + B_
            y = x_curve.spline(x)
            ax2.plot(x, y, 'o', color='cyan')

    fig.tight_layout()
    fig.subplots_adjust(top=0.85)
    plt.show()
    plt.pop()

def simple_plot_pair_curves(title, a_curve, x_curve):
    from DataUtils import get_in_folder
    in_folder = get_in_folder()

    with plt.Dp():
        fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
        fig.suptitle(title + " with %s" % in_folder, fontsize=20)
        simple_plot(ax1, a_curve, color="blue")
        simple_plot(ax2, x_curve, color="orange")
        fig.tight_layout()
        plt.show()

def mapping_proof(title, a_curve, x_curve, mapping):
    x = x_curve.x
    y = x_curve.y
    A, B = mapping
    uv_x = x*A + B

    with plt.Dp():
        fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
        fig.suptitle(title)

        simple_plot(ax1, a_curve, legend=False)
        simple_plot(ax2, x_curve)

        ax1.plot(uv_x, y, label="mapped")
        ax1.legend()

        fig.tight_layout()
        plt.show()

def make_unique_peak_info( old_list ):
    temp_list = sorted( old_list )

    new_list = []
    last_info = None
    for info in temp_list:
        if last_info is None or info[0] != last_info[0]:
            new_list.append( info )

        last_info = info

    return new_list

def make_unique_boundaries( old_list ):
    return sorted( list( set( old_list ) ) )

class ProxyCurve(ElutionCurve):
    def __init__( self, curve, peak_top_x, peak_info, boundaries, add_peak_info, add_boundaries, correct=False ):
        self.__dict__   = curve.__dict__.copy()
        self.logger     = logging.getLogger(__name__)
        self.peak_top_x = np.array(peak_top_x)
        self.peak_info  = peak_info
        self.boundaries = boundaries
        self.peak_slice = slice(0, len(curve.x))    # comptibility with PeakCurve
        self.emg_peaks = None
        if correct:
            self.correct_valley_bottm(debug=True)

    def correct_valley_bottm(self, debug=False):

        if debug:
            saved_boundaries = copy.deepcopy(self.boundaries)

        for k, b in enumerate(self.boundaries):
            start = max(0, b - 20)
            stop  = min(len(self.x), b + 20)
            y_ = self.spline(self.x[start:stop])
            ppp = np.argpartition(y_, 5)
            n = start + int(np.average(ppp[0:5+1]) + 0.5)
            if n != b:
                self.logger.info("corrected %dth valley bottm from %d to %d." % (k, b, n))

        if debug:
            fig = plt.figure()
            ax = fig.gca()
            ax.plot(saved_boundaries, self.spline(saved_boundaries), 'o', color='gray', alpha=0.2, markersize=10, label='saved_boundaries')
            simple_plot(ax, self, "correct_valley_bottm debug")
            fig.tight_layout()
            plt.show()

    def get_peak_region_width(self):
        # note that ElutionCurve does not currently have this method
        return self.peak_slice.stop - self.peak_slice.start

class MajorCurve(ElutionCurve):
    def __init__(self, curve):
        self.__dict__   = curve.__dict__.copy()
        self.logger     = logging.getLogger(__name__)
        self.peak_top_x = curve.major_peak_top_x
        self.peak_info  = curve.major_peak_info
        self.boundaries = curve.major_boundaries

class LessPeakCurve(ElutionCurve):
    def __init__(self, curve, num_less):
        self.__dict__   = curve.__dict__.copy()
        self.logger     = logging.getLogger(__name__)
        peak_height_ratios = [ curve.get_height_ratio(top) for top in curve.peak_top_x ]
        less_index = np.argpartition(peak_height_ratios, num_less)[:num_less]
        print('peak_height_ratios=', peak_height_ratios)
        print('less_index=', less_index)
        new_peak_top_x = []
        for k, top in enumerate(curve.peak_top_x):
            if k in less_index:
                continue
            new_peak_top_x.append(top)
        self.peak_top_x = np.array(new_peak_top_x)
        self.set_peak_dependent_info()

class MappedInfo(list):
    def __init__(self, info_list, curves=None):
        list.__init__(self, info_list)
        self._curves = curves
        best_params = info_list[0]
        if best_params is None:
            # this can happen in the preliminary recongnition as in 20171226
            pass
        else:
            self._A, self._B = best_params

    def xr_to_uv(self, xr_p):
        return self._A * xr_p + self._B

    def uv_to_xr(self, uv_p):
        return (uv_p - self._B)/self._A

def slice_length(s):
    return s.stop - s.start

class PeakMapper:
    def __init__(self, a_curve, x_curve, simply=False, pre_recog=None, debug=False, debug_detail=False ):
        self.logger = logging.getLogger(__name__)

        self.a_curve = a_curve
        self.x_curve = x_curve
        self.peak_mapping_only = get_setting('peak_mapping_only')
        self.feature_mapped = False
        self.reliable = True
        self.debug  = debug
        self.debug_detail = debug_detail

        fallback = False
        if not simply:
            try:
                sliced_pm_info = self.get_sliced_peak_mapper(a_curve, x_curve, debug=debug)
                self.pre_recog  = pre_recog
                self.make_it_compatible_from_sliced(sliced_pm_info, debug=debug)
            except Exception as exc:
                warnlog_exception(self.logger, "sliced peak mapping failed: ")
                simply = True
                fallback = True
                self.logger.warning("resorting to the simple treatment (i.e., without slicing)  due to the previous error.")

        if simply:
            if debug:
                simple_plot_pair_curves("PeakCurves (0)", self.a_curve, self.x_curve)

            try:
                self.pre_recog = None
                self.do_mapping(debug=debug)
                if fallback:
                    A, B =self.get_imroved_mapping(fallback=fallback)
                    self.mapped_info[0] = (A, B)
                    self.best_info[1] = (A, B)
            except Exception as exc:
                # as in 20201007_2
                from .ReducedCurve import make_reduced_curve
                self.logger.warning("resorting to curve reducing due to exception: %s", str(exc))
                curves = []
                for curve in self.a_curve, self.x_curve:
                    curves.append(make_reduced_curve(curve))
                self.do_mapping(reduced_curves=curves, debug=debug)

    def preliminary_check(self, a_curve, x_curve):
        a_frames = len(a_curve.x)
        x_frames = len(x_curve.x)
        frames_ratio = a_frames/x_frames
        almost_same_scale = abs(frames_ratio - 1) < 0.1

        peaktop_ratio = (a_curve.primary_peak_x/a_frames)/(x_curve.primary_peak_x/x_frames)
        proportinal_peaktop = abs(peaktop_ratio - 1) < 0.1

        a_peaks = len(a_curve.peak_info)
        x_peaks = len(x_curve.peak_info)
        same_few_num_peaks = a_peaks == x_peaks and x_peaks <= 2

        return Struct(  almost_same_scale=almost_same_scale,
                        frames_ratio=frames_ratio,
                        proportinal_peaktop=proportinal_peaktop,
                        a_peaks = a_peaks,
                        x_peaks = x_peaks,
                        same_few_num_peaks=same_few_num_peaks,
                        )

    def check_peak_existence(self, a_curve, x_curve, title, plot=True):
        for curve in a_curve, x_curve:
            if len(curve.peak_info) == 0:
                if plot:
                    with plt.Dp():
                        fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
                        fig.suptitle(title)
                        simple_plot(ax1, a_curve, color="blue")
                        simple_plot(ax2, x_curve, color="orange")
                        fig.tight_layout()
                        plt.show()
                assert False

    def get_sliced_peak_mapper(self, a_curve, x_curve, debug=False):
        self.check_peak_existence(a_curve, x_curve, "get_sliced_peak_mapper entry")

        check_info = self.preliminary_check(a_curve, x_curve)
        if debug:
            print("check_info=", check_info)
            simple_plot_pair_curves("get_sliced_peak_mapper (1)", a_curve, x_curve)

        peak_curve_ok = False
        if len(x_curve.x) >= NUM_FRAMES_BOUNADRY:
            try:
                a_curve_, a_slice_ = get_peak_curve_info(a_curve, debug=debug)
                x_curve_, x_slice_ = get_peak_curve_info(x_curve, debug=debug)
                try:
                    a_curve_, a_slice_, x_curve_, x_slice_ = self.improve_slices_if_possible(check_info, a_curve_, a_slice_, x_curve_, x_slice_)
                except:
                    warnlog_exception(self.logger, "improve_slices_if_possible failed: ")
                peak_curve_ok = True
            except:
                # as in 20201127_5
                warnlog_exception(self.logger, "get_sliced_peak_mapper(1)")
        elif len(a_curve.x) >= NUM_FRAMES_BOUNADRY*2:
            # as in 20201009_1
            try:
                a_curve_, a_slice_ = get_peak_curve_info(a_curve, debug=debug)
                x_curve_, x_slice_ = x_curve, None
                peak_curve_ok = True
            except:
                warnlog_exception(self.logger, "get_sliced_peak_mapper (2)")

        if not peak_curve_ok:
            # as in 20191114_3 where len(a_curve.x) >= 1000 while len(x_curve.x) < 1000
            # avoiding the inconvenience in get_peak_curve_info, which should better be investigated anyway
            a_curve_, a_slice_ = a_curve, None
            x_curve_, x_slice_ = x_curve, None

        if debug:
            simple_plot_pair_curves("get_sliced_peak_mapper (2)", a_curve_, x_curve_)

        self.check_peak_existence(a_curve_, x_curve_, "get_sliced_peak_mapper before recursive construction")

        pm = PeakMapper(a_curve_, x_curve_, simply=True, debug=debug)
        return Struct(pm=pm, a_slice_=a_slice_, x_slice_=x_slice_)

    def improve_slices_if_possible(self, check_info, a_curve_, a_slice_, x_curve_, x_slice_, debug=False):
        sliced_check_info = self.preliminary_check(a_curve_, x_curve_)
        # print("------------------- check_info=", check_info)
        # print("------------------- sliced_check_info=", sliced_check_info)
        if sliced_check_info.a_peaks > 2 and sliced_check_info.x_peaks >= 2:
            # as in 20220716/FER_OA_302
            return a_curve_, a_slice_, x_curve_, x_slice_

        if check_info.almost_same_scale:
            pass
        else:
            if check_info.proportinal_peaktop:
                pass
            else:
                ok = True
                if check_info.same_few_num_peaks:
                    if not sliced_check_info.same_few_num_peaks:
                        # as in 20200121_3
                        ok = False
                if ok:
                    return a_curve_, a_slice_, x_curve_, x_slice_
                else:
                    pass

        ratio = slice_length(a_slice_)/(check_info.frames_ratio * slice_length(x_slice_))
        if abs(ratio - 1) < 0.1:
            return a_curve_, a_slice_, x_curve_, x_slice_

        """
        try to improve the slice in cases like 20200121_2 or 20200630_11
        """

        if not check_info.same_few_num_peaks:
            # like 20200630_11
            # note that unmatching peaks will be removed
            num_ratio = len(a_curve_.peak_info)/len(x_curve_.peak_info)
            # assert (ratio - 1)*(num_ratio - 1) > 0      # i.e. of the same sign
            curve_name = "x_curve" if num_ratio < 1 else "a_curve"
            self.logger.info("unmatching peaks will probably be removed from %s", curve_name)

        if ratio < 1:
            curve1 = self.a_curve
            curve1_ = a_curve_
            slice1_ = a_slice_
            curve2 = self.x_curve
            curve2_ = x_curve_
            slice2_ = x_slice_
            self.logger.info("improve_slices_if_possible: improving x_slice_=%s from a_slice_=%s for ratio=%.3g", str(x_slice_), str(a_slice_), ratio)
        else:
            curve1 = self.x_curve
            curve1_ = x_curve_
            slice1_ = x_slice_
            curve2 = self.a_curve
            curve2_ = a_curve_
            slice2_ = a_slice_
            self.logger.info("improve_slices_if_possible: improving a_slice_=%s from x_slice_=%s for ratio=%.3g", str(a_slice_), str(x_slice_), ratio)

        sm = SimplestMapper(curve1, curve2)
        slice_ = slice(sm.map_1to2(slice1_.start), sm.map_1to2(slice1_.stop))
        ecurve = ElutionCurve(curve2.y[slice_])
        ecurve.get_default_editor_ranges()

        if debug:
            with plt.Dp():
                fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
                fig.suptitle("improve_slices_if_possible result")
                simple_plot(ax1, curve1)
                simple_plot(ax2, curve2, legend=False)
                ax2.plot(slice_.start+ecurve.x, ecurve.y, ":", label="with improved slice")
                ax2.legend()
                fig.tight_layout()
                plt.show()

        if ratio < 1:
            return curve1_, slice1_, ecurve, slice_
        else:
            return ecurve, slice_, curve1_, slice1_

    def check_minimum_mapping_quality(self, a_curve, x_curve, where_text):
        x = [0]
        y = [0]
        for info1, info2 in zip(a_curve.peak_info, x_curve.peak_info):
            x.append(info2[1])
            y.append(info1[1])
        x.append(len(x_curve.x))
        y.append(len(a_curve.x))

        correlation = pearsonr(x, y)[0]
        print(where_text, "------------------ correlation=", correlation)

    def do_mapping(self, reduced_curves=None, debug=False):

        if reduced_curves is None:
            a_curve = self.a_curve
            x_curve = self.x_curve
        else:
            a_curve, x_curve = reduced_curves

        if debug:
            simple_plot_pair_curves("before mapping", a_curve, x_curve)

        if self.debug:
            from CallStack import CallStack
            cstack = CallStack()
            print('PeakMapper call stack=', cstack)
            debug_plot( a_curve, x_curve, "before" )

        A, B = self.get_simplest_params(a_curve, x_curve)

        if debug:
            import Mapping.PeakMappingSolver
            from importlib import reload
            reload(Mapping.PeakMappingSolver)
        from .PeakMappingSolver import PeakMappingSolver
        pms = PeakMappingSolver(a_curve, x_curve, (A, B), debug=debug)
        uv_mapper, xr_mapper = pms.get_opt_mappers()
        try:
            better_a_curve = self.make_better_curve(a_curve, x_curve, xr_mapper, (A, B), "better_a_curve", debug=debug)
            better_x_curve = self.make_better_curve(x_curve, a_curve, uv_mapper, (1/A, -B/A), "better_x_curve", debug=debug)
            curves = [better_a_curve, better_x_curve]
        except Exception as exc:
            warnlog_exception(self.logger, "making thinner curves due to ")
            thinner_a_curve = self.make_thinner_curve(a_curve, uv_mapper)
            thinner_x_curve = self.make_thinner_curve(x_curve, xr_mapper)
            curves = [thinner_a_curve, thinner_x_curve]

        if self.debug:
            debug_plot(*curves, "after")

        self.check_peak_existence(*curves, "do_mapping result", plot=False)
        self.check_minimum_mapping_quality(*curves, "do_mapping result")

        self.mapped_curves = curves

        if debug:
            simple_plot_pair_curves("after mapping", *curves)

    def make_it_compatible_from_sliced(self, sliced_pm_info, debug=False):
        pm = sliced_pm_info.pm
        a_slice_ = sliced_pm_info.a_slice_
        a_start = 0 if a_slice_ is None else a_slice_.start
        x_slice_ = sliced_pm_info.x_slice_
        x_start = 0 if x_slice_ is None else x_slice_.start
        mapped_curves = pm.get_mapped_curves()
        # A, B = pm.best_info[1]
        A, B = pm.get_imroved_mapping()

        B_ = a_start + A*(-x_start) + B

        if debug:
            simple_plot_pair_curves("make_it_compatible_from_sliced", *mapped_curves)

        # None's will not be used
        self.best_info = [pm.best_info[0], (A,B_), None, None, None]
        self.mapped_info = MappedInfo([(A, B_), [], None], [self.a_curve, self.x_curve])
        curves = []
        for curve, mp_curve, slice_ in [(self.a_curve, mapped_curves[0], a_slice_),
                                        (self.x_curve, mapped_curves[1], x_slice_) ]:
            if slice_ is None:
                pkcurve = mp_curve
            else:
                pkcurve = PeakCurve(curve, pkslice_info=(mp_curve, slice_))
                pkcurve.get_emg_peaks()

                """
                try:
                    pkcurve = PeakCurve(curve, pkslice_info=(mp_curve, slice_))
                    pkcurve.get_emg_peaks()
                except:
                    warnlog_exception(self.logger, "make_it_compatible_from_sliced: ")
                    pkcurve = curve
                """

            curves.append(pkcurve)

        self.check_minimum_mapping_quality(*curves, "make_it_compatible_from_sliced result")
        self.mapped_curves = curves

        if debug:
            simple_plot_pair_curves("after make_it_compatible_from_sliced", *curves)

    def get_simplest_params(self, curve1, curve2):
        max_score, max_params, max_indeces, max_correl, max_simil, max_info = self.get_best_mapping_params(curve1, curve2, peak_mapping_only=False)
        B, A = max_params

        if len(curve1.peak_info) < 3 or len(curve2.peak_info) < 3:
            a, b = self.compare_primary_peak_mapping_params(curve1, curve2, (A, B))
        else:
            # avoid wornsening as with 202300705 (BSA)
            a, b = A, B

        # for backward compatibility begin
        self.best_info = [max_score, (a, b), max_indeces, max_correl, max_simil]
        self.mapped_info = MappedInfo([(a, b), [max_correl, max_simil], max_info], [self.a_curve, self.x_curve])
        # for backward compatibility end

        return a, b

    def get_mapped_curves( self ):
        return self.mapped_curves

    def find_best_mapping( self, curve1, curve2, check_adequacy=True, features=True):

        curve1_p, curve2_p, best_info_p, mapped_info_p = self.find_best_mapping_impl(curve1, curve2, False, check_adequacy, debug=False)

        if features and not self.peak_mapping_only:
            curve1_f, curve2_f, best_info_f, mapped_info_f = self.find_best_mapping_impl(curve1_p, curve2_p, True, check_adequacy, mapped_info=mapped_info_p, debug=False)

            if best_info_f[0] > best_info_p[0]:
                best_info = best_info_f
                mapped_info = mapped_info_f
                ret_curves = [curve1_f, curve2_f]
                self.feature_mapped = True
                self.logger.info("features matching has been successful: %g > %g" % (best_info_f[0],  best_info_p[0]))
            else:
                best_info = best_info_p
                mapped_info = mapped_info_p
                ret_curves = [curve1_p, curve2_p]

                if self.pre_recog is not None:
                    mapped_info = self.pre_recog.get_restricted_mapped_info()
                    curve1_f, curve2_f, best_info_f, mapped_info_f = self.find_best_mapping_impl(curve1_p, curve2_p, True, check_adequacy, mapped_info=mapped_info, debug=False)
                    if best_info_f[0] > best_info_p[0]:
                        best_info = best_info_f
                        mapped_info = mapped_info_f
                        ret_curves = [curve1_f, curve2_f]
                        self.feature_mapped = True

                if self.feature_mapped:
                    self.logger.info("features matching has been successful using pre-recog: %g > %g" % (best_info_f[0],  best_info_p[0]))
                else:
                    self.logger.warning("features matching has been worse: %g <= %g" % (best_info_f[0], best_info_p[0]))
        else:
            best_info = best_info_p
            mapped_info = mapped_info_p
            ret_curves = [curve1_p, curve2_p]

        self.mapped_info = mapped_info
        self.best_info = best_info

        return ret_curves

    def evaluate_mapping_impl( self, size1, size2, curve1, peaks1, index1, curve2, peaks2, index2, use_ends=True, debug=False):
        if use_ends and (len(peaks1) < 2 or len(peaks2) < 2):
            y   = [ 0, size1-1 ]
            x   = [ 0, size2-1 ]
            w   = [ 1, 1 ]
        else:
            y   = []
            x   = []
            w   = []

        for p1, p2 in zip( peaks1[index1], peaks2[index2] ):
            y.append( p1 )
            x.append( p2 )
            w.append( 10 )

        if not use_ends:
            for rec1, rec2 in zip(curve1.peak_info, curve2.peak_info):
                for i in [0, 2]:
                    y.append( rec1[i] )
                    x.append( rec2[i] )
                    w.append( 5 )

        X   = sm.add_constant(x)
        mod = sm.WLS( y, X, weights=w )
        res = mod.fit()

        if False:
            B, A = res.params
            mp_x = curve2.x*A + B
            mp_y = curve2.spline(mp_x)/curve2.max_y*curve1.max_y
            rmsd = np.sqrt(np.average((curve1.spline(mp_x) - mp_y)**2))

        if debug:
            with plt.Dp():
                fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
                fig.suptitle("evaluate_mapping_impl")
                simple_plot(ax1, curve1, color="blue")
                ax1.plot(mp_x, mp_y, ":")
                simple_plot(ax2, curve2, color="orange")
                fig.tight_layout()
                plt.show()

        correl, simil = self.compute_mapped_curve_simimarity(res.params, x, y, w, [index1, index2], curve1, curve2, debug=False )

        return correl, simil, res.params, [x, y, w]

    def evaluate_mapping( self, size1, size2, curve1, peaks1, index1, curve2, peaks2, index2):

        first_result = self.evaluate_mapping_impl(size1, size2, curve1, peaks1, index1, curve2, peaks2, index2)
        correl, simil, params, info = first_result
        first_score = correl + simil
        # print('first_score=', first_score, (correl, simil) )
        if len(curve1.peak_info) == 1 and len(curve2.peak_info) == 1 and simil < 0.9:
            try:
                second_result = self.evaluate_mapping_impl(size1, size2, curve1, peaks1, index1, curve2, peaks2, index2, use_ends=False)
                correl, simil, params, info = second_result
                second_score = correl + simil
                # print('second_score=', second_score, (correl, simil) )
            except:
                second_score = 0
            score_ratio = second_score/first_score
            print('score_ratio=', score_ratio)
            if score_ratio > ADOPTABLE_SCORE_RATIO:
                # there should be enough superiority to adopt the non-use_ends result
                self.logger.info("the 2nd result scored %.3g has been adopted over the 1st result scored %.3g." % (second_score, first_score) )
                ret_result = second_result
            else:
                ret_result = first_result
        else:
            ret_result = first_result

        return ret_result

    def evaluate_mapping_features( self, curve1, points1, weights1, index1, curve2, points2, weights2, index2):
        y = []
        x = []
        w = []

        assert_later = False

        inappropriate = False
        for p1, w1, p2, w2 in zip( points1[index1], weights1[index1], points2[index2], weights2[index2] ):
            y.append( p1 )
            x.append( p2 )
            if assert_later:
                if abs(w1 - w2) > 1e-6:
                    inappropriate = True
            else:
                # debug_plot(curve1, curve2, "evaluate_mapping_features")
                assert abs(w1 - w2) < 1e-6
            w.append( max(w1, w2) )

        X   = sm.add_constant(x)
        mod = sm.WLS( y, X, weights=w )
        res = mod.fit()

        correl, simil = self.compute_mapped_curve_simimarity(res.params, x, y, w, [index1, index2], curve1, curve2, features=True, debug=False )

        if assert_later:
            if inappropriate:
                assert False
                # pass

        return correl, simil, res.params, [x, y, w]

    def compute_mapped_curve_simimarity(self, params, x, y, w, indeces, curve1, curve2, features=False, nodebug=False, debug=False):
        intercept, slope = params

        correlation = pearsonr(x, y)[0]

        j_ = np.array([0, len(curve1.x)])
        i_ = 1/slope*j_ - intercept/slope
        i_start = int( min(0, i_[0]) )
        i_stop = int( max(len(curve2.x), i_[-1]))

        i = np.arange(i_start, i_stop)
        j = slope * i + intercept

        source_y = curve2.spline(i)
        scale = curve2.height/curve1.height
        mapped_y_ = curve1.spline(j) * scale

        def obj_func(params):
            d = np.sum(np.abs(mapped_y_ + params[0] - source_y))
            return d

        # allow any vertical shift and minimize the gap area
        result = minimize(obj_func, np.zeros(1))

        diff_area = obj_func(result.x)
        mapped_y = mapped_y_ + result.x[0]

        whole_area = curve2.height * len(source_y)
        similarity = abs(1 -  diff_area/whole_area)

        if debug:
        # if True:
            from matplotlib.patches import Polygon
            from DataUtils import get_in_folder

            in_folder = get_in_folder()
            markersize = 10

            plt.push()
            fig, axes = plt.subplots(nrows=1, ncols=3, figsize=(21,7))
            ax1, ax2, ax3 = axes
            ax1t = ax1.twinx()

            fig.suptitle("Mapping Explained for %s: indeces=%s" % (in_folder, str(indeces)), fontsize=20)
            ax1.set_title("Data Curves", fontsize=16)
            ax2.set_title("Elution No Mapping", fontsize=16)
            ax3.set_title("Mapped Curves", fontsize=16) 

            c_colors = ['blue', 'orange']
            p_colors = ['cyan', 'orange']

            ax1.plot(j, curve1.spline(j), color=c_colors[0])
            ax1t.plot(i, source_y, color=c_colors[1])

            ax2.plot(i, j, color='green')
            xmin2, xmax2 = ax2.get_xlim()
            ymin2, ymax2 = ax2.get_ylim()

            spanx = xmax2 - xmin2
            spany = ymax2 - ymin2
            xmin_ = -spanx*0.1
            ymin_ = -spany*0.1

            ax2.plot(x, [ymin_]*len(x), 'o', markersize=markersize, color=p_colors[1])
            ax2.plot([xmin_]*len(y), y, 'o', markersize=markersize, color=p_colors[0])

            for px, py, pw in zip(x, y, w):
                ax1.plot(py, curve1.spline(py), 'o', markersize=markersize, color=p_colors[0])
                ax1t.plot(px, curve2.spline(px), 'o', markersize=markersize, color=p_colors[1])
                ax2.plot(px, py, 'o', markersize=markersize, color='lime')
                py_ = slope * px + intercept
                ax3.plot(px, curve1.spline(py_) * scale, 'o', markersize=markersize, color=p_colors[0])
                ax3.plot(px, curve2.spline(px), 'o', markersize=markersize, color=p_colors[1])
                shrinky = spany*0.02/(py - ymin_)
                ax2.annotate("", xy=(px, py), xytext=(px, ymin_), alpha=0.5,
                            arrowprops=dict(arrowstyle='->', ls='dashed', lw=1.5, color=c_colors[1], shrinkA=10, shrinkB=10),
                            )
                shrinkx = spanx*0.02/(px - xmin_)
                ax2.annotate("", xy=(xmin_, py), xytext=(px, py), alpha=0.5,
                            arrowprops=dict(arrowstyle='->', ls='dashed', lw=1.5, color=c_colors[0], shrinkA=10, shrinkB=10),
                            )

                if pw > 0.9:
                    ax1.plot(py, curve1.spline(py), 'o', markersize=5, color='red')
                    ax1t.plot(px, curve2.spline(px), 'o', markersize=5, color='red')
                    ax3.plot(px, curve1.spline(py_) * scale, 'o', markersize=5, color='red')
                    ax3.plot(px, curve2.spline(px), 'o', markersize=5, color='red')

            ax3.plot(i, source_y, color=c_colors[1])
            ax3.plot(i, mapped_y, color=c_colors[0])
            ax3.plot(i, mapped_y_, ':', color=c_colors[0])

            poly_points = list(zip( i, source_y )) + list( reversed( list( zip( i, mapped_y ) ) ) )
            diff_poly   = Polygon( poly_points, alpha=0.1 )
            ax3.add_patch(diff_poly)

            xmin, xmax = ax2.get_xlim()
            ymin, ymax = ax2.get_ylim()
            tx = (xmin + xmax)/2
            ty = (ymin + ymax)/2
            ax2.text(tx, ty, "Linearity=%.3g" % correlation, ha='center', va='center', fontsize=50, alpha=0.1)

            xmin, xmax = ax3.get_xlim()
            ymin, ymax = ax3.get_ylim()
            tx = (xmin + xmax)/2
            ty = (ymin + ymax)/2
            ax3.text(tx, ty, "Similarity=%.3g" % similarity, ha='center', va='center', fontsize=50, alpha=0.1)

            fig.tight_layout()
            fig.subplots_adjust(top=0.85)
            plt.show()
            plt.pop()

        if False and not nodebug and self.debug_detail or debug:
            from scipy.interpolate import UnivariateSpline
            from matplotlib.patches import Polygon

            colors = ['blue', 'orange']

            size1   = len(curve1.y)
            size2   = len(curve2.y)

            if features:
                points1, weights1 = curve1.feature.get_points()
                points2, weights2 = curve2.feature.get_points()
                peaks1  = points1
                peaks2  = points2
                points_label = 'points'
                weight_boundary = 0.75
            else:
                peaks1  = np.array( [ info[1] for info in curve1.peak_info ] )
                peaks2  = np.array( [ info[1] for info in curve2.peak_info ] )
                points_label = 'peaks'
                weight_boundary = 5

            unmapped_index = []
            for k in range(len(peaks1)):
                if k not in indeces:
                    unmapped_index.append(k)

            mapped_spline = UnivariateSpline(i, mapped_y, s=0, ext=3)
            mapped_peaks = 1/slope*peaks1 - intercept/slope

            fig = plt.figure(figsize=(14,7))
            fig.suptitle("Debug: ac_vector: indeces=" + str(indeces))

            ax1 = fig.add_subplot(121)
            ax2 = fig.add_subplot(122)
            ax1.set_title("Index Mapping") 
            ax2.set_title("Mapped Curves") 

            x_  = [ 0, size2-1 ]
            y_  = [ slope*z + intercept  for z in x_ ]
            ax1.plot( x_, y_, color='green' )

            for px, py, pw in zip(x, y, w):
                # color = 'red' if pw > weight_boundary else 'black'
                color = 'red'
                msize = 10 if pw > weight_boundary else 5
                ax1.plot(px, py, 'o', color=color, markersize=msize)

            ax2.plot(i, source_y, color=colors[1], label='source')
            ax2.plot(i, mapped_y, color=colors[0], label='mapped')
            ax2.plot(i, mapped_y_, ':', color=colors[0], label='mapped')

            if len(unmapped_index) > 0:
                mapped_peaks_ = mapped_peaks[unmapped_index]
                ax2.plot(mapped_peaks_, mapped_spline(mapped_peaks_), 'o', color='yellow', markersize=10, label='unmapped ' + points_label)

            """
            TO BE FIXED:
            there was an error occured in 20170511
            IndexError: index 1 is out of bounds for axis 0 with size 1
            """
            print('mapped_peaks=', mapped_peaks, 'indeces=', indeces)
            mapped_peaks_ = mapped_peaks[indeces[0]]
            ax2.plot(mapped_peaks_, mapped_spline(mapped_peaks_), 'o', color='red', markersize=10, label='mapped ' + points_label)

            ax2.plot(peaks2, curve2.spline(peaks2), 'o', color='pink', markersize=10, label='source ' + points_label)

            poly_points = list(zip( i, source_y )) + list( reversed( list( zip( i, mapped_y ) ) ) )
            diff_poly   = Polygon( poly_points, alpha=0.2 )
            ax2.add_patch(diff_poly)

            xmin, xmax = ax1.get_xlim()
            ymin, ymax = ax1.get_ylim()
            tx = (xmin + xmax)/2
            ty = (ymin + ymax)/2
            ax1.text(tx, ty, "Linearity=%.3g" % correlation, ha='center', va='center', fontsize=60, alpha=0.2)

            xmin, xmax = ax2.get_xlim()
            ymin, ymax = ax2.get_ylim()
            tx = (xmin + xmax)/2
            ty = (ymin + ymax)/2
            ax2.text(tx, ty, "Similarity=%.3g" % similarity, ha='center', va='center', fontsize=60, alpha=0.2)

            ax2.legend()
            fig.tight_layout()
            fig.subplots_adjust(top=0.9)
            plt.show()

        return correlation, similarity

    def get_best_mapping_params(self, curve1, curve2, peak_mapping_only=False):
        if peak_mapping_only:
            peaks1  = np.array( [ info[1] for info in curve1.get_major_peak_info(ratio=0.3) ] )     # ratio=0.1 is too small for 20190221_2
            peaks2  = np.array( [ info[1] for info in curve2.get_major_peak_info(ratio=0.3) ] )     # 
        else:
            peaks1  = np.array( [ info[1] for info in curve1.peak_info ] )
            peaks2  = np.array( [ info[1] for info in curve2.peak_info ] )

        size1   = len(curve1.y)
        size2   = len(curve2.y)
        num_peaks1 = len(peaks1)
        num_peaks2 = len(peaks2)

        greater_num = max(num_peaks1, num_peaks2)
        less_num = min(num_peaks1, num_peaks2)

        greater_index = list(range(greater_num))
        full_index1 = list(range(num_peaks1))
        full_index2 = list(range(num_peaks2))

        max_score   = None
        max_params  = None
        max_indeces = None
        max_correl  = None
        max_simil   = None
        max_info    = None

        for c in combinations(greater_index, less_num):
            index = list(c)
            if num_peaks1 >= num_peaks2:
                index1 = index
                index2 = full_index2
            else:
                index1 = full_index1
                index2 = index
            assert len(index1) == len(index2)
            correl, simil, params, info = self.evaluate_mapping( size1, size2, curve1, peaks1, index1, curve2, peaks2, index2 )
            score = correl + simil
            if max_score is None or score > max_score:
                max_score   = score
                max_params  = params
                max_indeces = (index1, index2)
                max_correl  = correl
                max_simil   = simil
                max_info    = info

        return max_score, max_params, max_indeces, max_correl, max_simil, max_info

    def get_best_mapping_features_params( self, curve1, curve2, mapped_info ):

        major_peak_info1 = curve1.get_major_peak_info()
        major_peak_info2 = curve2.get_major_peak_info()

        adopted_peak_info1 = []
        adopted_peak_info2 = []
        best_params = mapped_info[0]
        A, B = best_params
        A_, B_ = 1/A, -B/A

        size2 = len(curve2.x)
        for rec1 in major_peak_info1:
            top1 = rec1[1]
            mapped_top1 = top1*A_ + B_
            for rec2 in major_peak_info2:
                top2 = rec2[1]
                if abs(mapped_top1 - top2)/size2 < 0.05:
                    adopted_peak_info1.append(rec1)
                    adopted_peak_info2.append(rec2)
                    break

        points1, weights1 = curve1.feature.get_points(adopted_peak_info1)
        points2, weights2 = curve2.feature.get_points(adopted_peak_info2)

        num_points1 = len(points1)
        num_points2 = len(points2)

        diff_num = abs(num_points1 - num_points2)

        if REDUCE_WHEN_INFEASIBLE and diff_num > FEASIBLE_DIFF_NUM:
            """
            not yet usable for 20170426
            """
            print('num_points1=', num_points1, 'num_points2=', num_points2)

            A_, B_ = 1/A, -B/A
            self.logger.warning("reducing points: %d vs. %d." % (num_points1, num_points2))

            def reduce_num_points(scale, shift, more_points, more_weights, less_points, less_weights):
                np1 = len(more_points)
                np2 = len(less_points)
                min_distances = [np.min( np.abs( scale*p+shift - less_points) ) for p in more_points]
                k = np1 - (np2 + FEASIBLE_DIFF_NUM)
                ppp = np.argpartition(min_distances, k)
                adopted_index = ppp[k:]
                print('adopted_index=', adopted_index)
                ret_points = more_points[adopted_index]
                ret_weights = more_weights[adopted_index]
                return ret_points, ret_weights

            if num_points1 > num_points2:
                points1, weights1 = reduce_num_points(A_, B_, points1, weights1, points2, weights2)
                num_points1 = len(points1)
            else:
                points2, weights2 = reduce_num_points(A, B, points2, weights2, points1, weights1)
                num_points2 = len(points2)

            diff_num = num_points1 - num_points2
            self.logger.warning("reduced points: %d vs. %d." % (num_points1, num_points2))

            if False:
                names = ['UV', 'Xray']
                colors = ['blue', 'orange']

                plt.push()
                fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(14,7))
                ax1, ax2 = axes
                simple_plot(ax1, curve1, names[0], color=colors[0], legend=False)
                simple_plot(ax2, curve2, names[1], color=colors[1], legend=False)
                ax1.plot(points1, curve1.spline(points1), 'o', color='yellow')
                ax2.plot(points2, curve2.spline(points2), 'o', color='yellow')

                mapped_points = A_*points1+B_
                ax2.plot(mapped_points, curve2.spline(mapped_points), 'o', color='cyan')
                fig.tight_layout()
                plt.show()
                plt.pop()

        assert diff_num <= FEASIBLE_DIFF_NUM

        greater_num = max(num_points1, num_points2)
        less_num = min(num_points1, num_points2)

        greater_index = list(range(greater_num))
        full_index1 = list(range(num_points1))
        full_index2 = list(range(num_points2))

        max_score   = None
        max_params  = None
        max_indeces = None
        max_correl  = None
        max_simil   = None
        max_info    = None

        for c in combinations(greater_index, less_num):
            index = list(c)
            if num_points1 >= num_points2:
                index1 = index
                index2 = full_index2
            else:
                index1 = full_index1
                index2 = index
            try:
                correl, simil, params, info = self.evaluate_mapping_features(curve1, points1, weights1, index1, curve2, points2, weights2, index2)
            except AssertionError:
                # invalid combination
                if False:
                    log_exception(self.logger, "evaluate_mapping_features failure: ")
                continue

            score = correl + simil
            if max_score is None or score > max_score:
                max_score   = score
                max_params  = params
                max_indeces = (index1, index2)
                max_correl  = correl
                max_simil   = simil
                max_info    = info

        pre_recog_covered = False
        if max_score is None:
            if self.pre_recog is None:
                assert False
            else:
                best_params, max_score_pair, max_info = self.pre_recog.get_restricted_mapped_info()
                a, b = best_params
                max_params = np.array([b, a])
                x, y, w = max_info
                correl, simil = self.compute_mapped_curve_simimarity(max_params, x, y, w, None, curve1, curve2, features=True, nodebug=True)
                max_score = correl + simil
                # max_score = 2
                max_corel = correl
                max_simil = simil
                pre_recog_covered = True
                self.logger.warning("to cover the failure, feature matching params are replaced with the pre-recognized params.")

        return max_score, max_params, max_indeces, max_correl, max_simil, max_info

    def find_best_mapping_impl(self, curve1, curve2, features, check_adequacy=True, mapped_info=None, debug=False):

        num_peaks_1 = len(curve1.peak_info)
        num_peaks_2 = len(curve2.peak_info)
        assert num_peaks_1 == len(curve1.boundaries) + 1
        assert num_peaks_2 == len(curve2.boundaries) + 1

        debug = False
        # if debug and self.debug:
        if debug:
            def best_debug_plot(title, curve1, curve2, feature_info=None, orig_curves=None):
                from DataUtils import get_in_folder
                reversed_=False
                plt.push()
                fig, axes = plt.subplots( figsize=(21, 8), nrows=1, ncols=2 )
                fig.suptitle(title + ' for ' + get_in_folder(), fontsize=30)
                ax1, ax2 = axes
                if reversed_:
                    c1, c2 = curve2, curve1
                else:
                    c1, c2 = curve1, curve2

                if orig_curves is None:
                    c1_orig = c1
                    c2_orig = c2
                else:
                    if reversed_:
                        c2_orig, c1_orig = orig_curves
                    else:
                        c1_orig, c2_orig = orig_curves

                points1, weights1 = c1_orig.feature.get_points()
                points2, weights2 = c2_orig.feature.get_points()
                ax1.plot(points1, c1_orig.spline(points1), 'o', color='yellow', markersize=20, label='original feature points')
                ax2.plot(points2, c2_orig.spline(points2), 'o', color='yellow', markersize=20, label='original feature points')

                if orig_curves is not None and feature_info is not None:
                    points1, weights1 = c1.feature.get_points()
                    points2, weights2 = c2.feature.get_points()
                    alt_index, alt_params = feature_info
                    if len(points1) < len(points2):
                        index1 = list(range(len(points1)))
                        index2 = alt_index
                    else:
                        index1 = alt_index
                        index2 = list(range(len(points2)))
                    points1_ = points1[index1]
                    points2_ = points2[index2]
                    ax1.plot(points1_, c1.spline(points1_), 'o', color='cyan', markersize=15, label='mapped feature points')
                    ax2.plot(points2_, c2.spline(points2_), 'o', color='cyan', markersize=15, label='mapped feature points')

                    # mapped_points = points_[alt_index]
                    # ax_.plot(mapped_points, c_.spline(mapped_points), color='cyan', markersize=10, label='mapped points')

                simple_plot(ax1, c1, "UV", color='blue', legend=False)
                simple_plot(ax2, c2, "Xray", color='orange', legend=False)

                ax1.legend(fontsize=16)
                ax2.legend(fontsize=16)
                fig.tight_layout()
                fig.subplots_adjust(top=0.85)
                plt.show()
                plt.pop()
            best_debug_plot("find_best_mapping begin", curve1, curve2)

        feature_info = None

        if features:
            try:
                max_score, max_params, max_indeces, max_correl, max_simil, max_info = self.get_best_mapping_features_params(curve1, curve2, mapped_info=mapped_info)
                feature_info = [ max_indeces, max_params ]
                B, A = max_params
                best_params = np.array([A, B])
                self.logger.info("get_best_mapping_features_params has been successful: A=%g, B=%g" % (A, B))
            except:
                self.logger.warning(" " + etb.last_lines())
                log_exception(self.logger, "get_best_mapping_features_params failed: ")
                max_score, best_params, max_indeces, max_correl, max_simil, max_info = 0, None, None, None, None, None
            curve1_, curve2_ = curve1, curve2
        else:
            index1 = list(range(num_peaks_1))
            max_score, max_params, max_indeces, max_correl, max_simil, max_info = self.get_best_mapping_params(curve1, curve2)
            # curve1_, curve2_, best_params = self.reconstruct_curves(curve1, curve2, max_params, max_indeces, max_correl, max_simil, check_adequacy)
            curve1_, curve2_, best_params = self.reconstruct_curves_simple(curve1, curve2, max_params, max_indeces,)

        best_info = [max_score, best_params, max_indeces, max_correl, max_simil]
        mapped_info = MappedInfo([best_params, [max_correl, max_simil], max_info], [self.a_curve, self.x_curve])

        # if self.debug and debug:
        if debug:
            best_debug_plot("find_best_mapping result", curve1_, curve2_, feature_info, orig_curves=[curve1, curve2])

        return curve1_, curve2_, best_info, mapped_info

    def reconstruct_curves_simple(self, curve1, curve2, max_params, max_indeces):
        B, A = max_params
        A_  = 1/A
        B_  = -B/A

        new_peak_top_x1 = []
        new_peak_info1 = []
        new_boundaries1 = []
        new_peak_top_x2 = []
        new_peak_info2 = []
        new_boundaries2 = []

        index1, index2 = max_indeces
        prev_pair = None
        for n, pair in enumerate(zip(index1, index2)):
            m, k = pair

            top_x1 = curve1.peak_top_x[m]
            peak_rec1 = curve1.peak_info[m]
            new_peak_top_x1.append(top_x1)
            new_peak_info1.append(peak_rec1)

            top_x2 = curve2.peak_top_x[k]
            peak_rec2 = curve2.peak_info[k]
            new_peak_top_x2.append(top_x2)
            new_peak_info2.append(peak_rec2)

            if n > 0:
                boundary1, boundary2 = self.find_valley_bottom_pair(curve1, curve2, prev_pair, m, k, A_, B_)
                new_boundaries1.append(boundary1)
                new_boundaries2.append(boundary2)

            prev_pair = pair

        assert len(new_peak_info1) == len(new_boundaries1) + 1
        assert len(new_peak_info2) == len(new_boundaries2) + 1
        assert len(new_peak_info1) == len(new_peak_info2)

        curve1_ =  ProxyCurve( curve1, new_peak_top_x1, new_peak_info1, new_boundaries1, [], [] )
        curve2_ =  ProxyCurve( curve2, new_peak_top_x2, new_peak_info2, new_boundaries2, [], [] )

        best_params = np.array([A, B])
        return curve1_, curve2_, best_params

    def find_valley_bottom_pair(self, curve1, curve2, prev_pair, m, k, A_, B_):
        prev_m, prev_k = prev_pair
        bc1_list = curve1.boundaries[prev_m:m]
        bc2_list = curve2.boundaries[prev_k:k]

        # select the best matching bottom pair
        min_dist = None
        min_j = None
        min_i = None
        for j, bc1 in enumerate(bc1_list):
            for i, bc2 in enumerate(bc2_list):
                dist = abs(bc1*A_ + B_ - bc2 )
                if min_dist is None or dist < min_dist:
                    min_dist = dist
                    min_j = j
                    min_i = i

        return bc1_list[min_j], bc2_list[min_i]

    def add_corresponding_peaks(self, curve1, curve2, A_, B_, debug_text=""):
        debug = True
        if debug:
            from matplotlib.patches import Rectangle
            from DataUtils import get_in_folder
            print('A_, B_=', A_, B_)
            plt.push()
            fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(14, 7))
            ax1, ax2 = axes
            fig.suptitle("add_corresponding_peaks entry: " + debug_text, fontsize=20)
            simple_plot(ax1, curve1, "curve1", color="blue")
            simple_plot(ax2, curve2, "curve2", color="orange")

            ymin, ymax = ax2.get_ylim()
            ax2.set_ylim(ymin, ymax)
            allow = len(curve2.x) * 0.05
            for top_x in curve2.peak_top_x:
                f = top_x - allow
                t = top_x + allow
                p = Rectangle(
                        (f, ymin),  # (x,y)
                        t - f,   # width
                        ymax - ymin,    # height
                        facecolor   = 'yellow',
                        alpha       = 0.3,
                    )
                ax2.add_patch(p)

            xmin, xmax = ax2.get_xlim()
            ax2t = ax2.twinx()
            ax2t.grid(False)
            ax2t.set_xlim(xmin, xmax)
            mapped_x = A_ * curve1.x  + B_
            ax2t.plot(mapped_x, curve1.y, ':', color='blue', label='mapped data')
            for top_x in curve1.peak_top_x:
                mx = A_ * top_x + B_
                my = curve1.spline(top_x)
                ax2t.plot(mx, my, 'o', color='yellow', label='mapped peak tops')

            ax2t.legend(bbox_to_anchor=(1, 0.9), loc='upper right')
            fig.tight_layout()
            plt.show()
            plt.pop()

        new_peak_top_x2 = []
        new_peak_info2 = []
        new_boundaries2 = []

        size2 = len(curve2.x)
        peaks2 = []
        start = 0
        stop = None
        last_peak_rec2 = None
        for m, top_x1 in enumerate(curve1.peak_top_x):
            mapped_top_x = A_*top_x1 + B_

            bottom_added = False
            if m > 0 and len(new_peak_top_x2) > 0:
                try:
                    bottom_x1 = curve1.boundaries[m-1]
                    mapped_bottom_x = A_*bottom_x1 + B_
                    bottom_x2 = self.find_counter_part_bottom(curve2, mapped_bottom_x, debug)
                    new_boundaries2.append(bottom_x2)
                    bottom_added = True
                except AssertionError:
                    pass

            try:
                mapped_peak_rec = [int(A_*i + B_ + 0.5) for i in curve1.peak_info[m]]
                top_x2, peak_rec2 = self.find_counter_part_peak(curve2, mapped_top_x, mapped_peak_rec, debug)
                new_peak_top_x2.append(top_x2)
                if last_peak_rec2 is not None:
                    if last_peak_rec2[2] >= peak_rec2[0]:
                        b = new_boundaries2[-1]
                        r = peak_rec2[0]
                        peak_rec2[0] = int((last_peak_rec2[2] + b)/2)
                        self.logger.warning("the left end modified from %d to %d in the %d-th range in add_corresponding_peaks", m, r, peak_rec2[0])
                new_peak_info2.append(peak_rec2)
                last_peak_rec2 = peak_rec2
            except AssertionError:
                warnlog_exception(self.logger, "add_corresponding_peaks: try peak")
                if bottom_added:
                    new_boundaries2.pop(-1)

            if debug:
                print("--------------------", [m], "new_peak_info2=", new_peak_info2)

        # assert len(new_peak_top_x2) == len(new_boundaries2) + 1

        if len(new_peak_top_x2) == len(new_boundaries2) + 1:
            curve2_ =  ProxyCurve( curve2, new_peak_top_x2, new_peak_info2, new_boundaries2, [], [] )
        else:
            curve2_ = curve2

        if debug:
            print('len(new_peak_top_x2)=', len(new_peak_top_x2))
            print('len(new_boundaries2)=', len(new_boundaries2))
            plt.push()
            fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(14, 7))
            ax1, ax2 = axes
            fig.suptitle("add_corresponding_peaks exit: " + debug_text)
            simple_plot(ax1, curve1, "curve1")
            simple_plot(ax2, curve2_, "curve2_")
            fig.tight_layout()
            plt.show()
            plt.pop()

        return curve2_

    def find_the_nearest_index(self, curve, vecx, px):
        """
        must choose the nearest since diff_ratio < 0.05 can detect multiple candidates
        """
        pr_width = curve.get_peak_region_width()

        min_ratio = None
        min_k = None
        for k, x_ in enumerate(vecx):
            diff_ratio = abs(px - x_)/pr_width
            if diff_ratio < 0.05:
                if min_ratio is None or diff_ratio < min_ratio:
                    min_k = k
                    min_ratio = diff_ratio

    def find_counter_part_peak(self, curve, ref_peak_x, ref_peak_rec, debug=False):
        k = self.find_the_nearest_index(curve, curve.peak_top_x, ref_peak_x)
        if k is not None:
            if debug:
                print('find_counter_part_peak: found', k)
            return curve.peak_top_x[k], curve.peak_info[k]

        delta_x = len(curve.x)*0.05
        start = int(ref_peak_x - delta_x)
        stop  = int(ref_peak_x + delta_x)

        n = find_rotated_extreme_arg(curve, start, stop, +1)

        # avoid minor peaks such as the point near 650 in 20170304 UV
        # default delta_x is too small for the 1st minor peak in 20160227
        assert curve.locally_tall_enough_evenif_rotated(n, delta_x=delta_x)

        if debug:
            print('-------------------- find_counter_part_peak: found rotated', n)

        return n, [ref_peak_rec[0], n, ref_peak_rec[2]]

    def find_counter_part_bottom(self, curve, ref_bottom_x, debug=False):
        k = self.find_the_nearest_index(curve, curve.boundaries, ref_bottom_x)
        if k is not None:
            return curve.boundaries[k]

        delta_x = len(curve.x)*0.05
        start = int(ref_bottom_x - delta_x)
        stop  = int(ref_bottom_x + delta_x)

        n = find_rotated_extreme_arg(curve, start, stop, -1)

        if debug:
            print('-------------------- find_counter_part_bottom: found rotated', n)

        return n

    def show_mapping_proof(self):
        a_curve_, x_curve_  = self.get_mapped_curves()
        mapping_proof("mapping proof", a_curve_, x_curve_, self.best_info[1])

    def make_better_curve(self, curve1, curve2, mapper, mapping, debug_text, debug=False):
        nomatch = mapper[mapper < 0]
        if len(nomatch) == 0:
            return curve1

        A, B = mapping

        new_peak_top_x = []
        new_peak_info = []
        new_boundaries = []

        max_y = curve1.max_y
        ptx1 = curve1.peak_top_x
        pif1 = curve1.peak_info
        bdy1 = curve1.boundaries
        ptx_ = curve2.peak_top_x*A + B
        bry_ = np.array(curve2.boundaries)*A + B

        if debug:
            print("mapper=", mapper)
            if debug_text.find("a_curve") >= 0:
                curve1_id = "UV"
                curve2_id = "XR"
            else:
                curve1_id = "XR"
                curve2_id = "UV"
            with plt.Dp():
                fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
                fig.suptitle("%s entry" % debug_text)
                ax1.set_title(curve1_id + " Elution")
                ax2.set_title(curve2_id + " Elution")
                simple_plot(ax1, curve1, legend=False)

                pty_ = curve1.spline(ptx_)
                ax1.plot(ptx_, pty_, "o", color="cyan", markersize=20, alpha=0.2, label="mapped peak")
                ax1.legend()

                simple_plot(ax2, curve2)
                fig.tight_layout()
                plt.show()

        ins = 0
        last_peak_info = None
        def adjusted_peak_info(temp_info):
            if last_peak_info is None or last_peak_info[2] < temp_info[0]:
                return temp_info

            ret_info = [last_peak_info[2]+1, *temp_info[1:]]
            assert ret_info[0] < ret_info[1]

            return ret_info

        try:
            for k, j in enumerate(mapper):
                if j >= 0:
                    continue
                x_ = ptx_[k]
                i = bisect_right(ptx1, x_)
                print([k], "i=", i)
                n = 0
                for p in range(ins, i):
                    new_peak_top_x.append(ptx1[p])
                    peak_info = adjusted_peak_info(pif1[p])
                    new_peak_info.append(peak_info)
                    last_peak_info = peak_info
                    if p > 0:
                        new_boundaries.append(bdy1[p-1])
                    n += 1
                ins += n
                top_x = ptx_[i]

                height_ratio = curve1.spline(top_x)/max_y
                # print([k], "-------------------- height_ratio=", height_ratio)
                if height_ratio < MINIMUM_HEIGHT_RATIO:
                    self.logger.info("peak with height_ratio=%.3g < %.3g will be discarded.", height_ratio, MINIMUM_HEIGHT_RATIO)
                    # eventually resorting to make_thinner_curve
                    assert False

                new_peak_top_x.append(top_x)
                peak_info = adjusted_peak_info([int(round(p*A + B)) for p in curve2.peak_info[i]])
                new_peak_info.append(peak_info)
                last_peak_info = last_peak_info
                if len(bry_) > 0:
                    new_boundaries.append(bry_[max(0, i-1)])    # task: prove max(0, i-1) is correct
        except AssertionError as exc:
            if debug:
                mx_ = curve2.x*A + B
                scale = curve1.max_y/curve2.max_y
                with plt.Dp():
                    fig, ax = plt.subplots()
                    ax.set_title("adjusted_peak_info AssertionError")
                    simple_plot(ax, curve1, legend=False)
                    ax.plot(mx_, scale*curve2.y, ":", label="mapped")
                    ax.plot(ptx_, scale*curve2.spline(curve2.peak_top_x), "o", markersize=20, color="cyan", alpha=0.2)
                    ax.legend()
                    fig.tight_layout()
                    plt.show()
            raise exc

        for p in range(ins, len(ptx1)):
            new_peak_top_x.append(ptx1[p])
            new_peak_info.append(pif1[p])
            if p > 0:
                new_boundaries.append(bdy1[p-1])

        curve1_ =  ProxyCurve( curve1, new_peak_top_x, new_peak_info, new_boundaries, [], [] )

        if debug:
            with plt.Dp():
                fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
                fig.suptitle("%s result" % debug_text)
                simple_plot(ax1, curve1_, legend=False)

                pty_ = curve1_.spline(ptx_)
                ax1.plot(ptx_, pty_, "o", color="cyan", markersize=20, alpha=0.2, label="mapped peak")
                ax1.legend()

                simple_plot(ax2, curve2)
                fig.tight_layout()
                plt.show()

        return curve1_

    def make_thinner_curve(self, curve, mapper):
        nomatch = mapper[mapper < 0]
        if len(nomatch) == 0:
            return curve

        new_peak_top_x = []
        new_peak_info = []
        new_boundaries = []

        ptx = curve.peak_top_x
        pif = curve.peak_info
        bdy = curve.boundaries

        n = 0
        for k, j in enumerate(mapper):
            if j < 0:
                # primary_peak must not be lost
                if k == curve.primary_peak_no:
                    raise AssertionError("primary peak lost")
                continue

            new_peak_top_x.append(ptx[k])
            new_peak_info.append(pif[k])
            if n > 0 and k > 0:
                new_boundaries.append(bdy[k-1])
            n += 1

        curve_ =  ProxyCurve( curve, new_peak_top_x, new_peak_info, new_boundaries, [], [] )
        return curve_

    def can_be_deemed_single_peak(self, curves):
        if len(curves[0].peak_info) == 1:
            return True

        num_peaks = []
        for curve in curves:
            n = 0
            max_y = curve.max_y
            for info in curve.peak_info:
                ratio = curve.spline(info[1])/max_y
                if ratio > DEEM_SINGLE_PEAK_RATIO:
                    n += 1
            num_peaks.append(n)
        return np.min(num_peaks) == 1

    def get_imroved_mapping(self, fallback=False, debug=False):
        from LPM import get_corrected   # moved due to ImportError: ... (most likely due to a circular import)

        if debug:
            from DebugUtils import show_call_stack
            show_call_stack("----: ", indented_only=True)            

        curves = self.mapped_curves
        A, B = self.best_info[1]

        if not self.can_be_deemed_single_peak(curves):
            return A, B

        max_simil = self.best_info[4]
        if len(curves[0].peak_info) == 1:
            uv_px = curves[0].peak_info[0][1]
            mp_px = A * curves[1].peak_info[0][1] + B
            dx = uv_px - mp_px
            self.logger.info("an adjustment %.3g of peak top deviation will be applied.", dx)
        else:
            dx = 0

        B += dx
        if max_simil < ACCEPTABLE_SIMILARITY:
            in_fallback = " in fallback" if fallback else ""
            self.logger.info("trying to improve to mapping due to poor similarity %.3g < %.3g%s.", max_simil, ACCEPTABLE_SIMILARITY, in_fallback)
        else:
            return A, B

        x1 = curves[0].x
        y1 = get_corrected(curves[0].y)
        i1 = curves[0].primary_peak_i
        x2 = curves[1].x
        y2 = get_corrected(curves[1].y)
        i2 = curves[1].primary_peak_i

        spline = UnivariateSpline(x1, y1, s=0, ext=3)
        scale = y1[i1]/y2[i2]

        def objective(p):
            a, b = p
            x = a*x2 + b
            return np.sum((spline(x) - scale*y2)**2)

        ret = minimize(objective, (A, B), method="Nelder-Mead")
        a, b = ret.x

        mapped_ends = [ex*a + b for ex in x2[[0, -1]]]
        ratio = (mapped_ends[1] - mapped_ends[0])/(x1[-1] - x1[0])
        if abs(ratio - 1) > ACCEPTABLE_ENDS_RATIO_DEV:
            # as in 20200123_4
            self.logger.info("mapping (%.3g, %.3g) is discarded due to ratio=%.3g", a, b, ratio)
            a, b = A, B

        if debug:
            print("fallback=", fallback)
            spline2 = UnivariateSpline(x2, y2, s=0, ext=3)
            with plt.Dp():
                fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
                fig.suptitle("get_imroved_mapping")
                ax1.set_title("Uncertain Mapping")
                ax1.plot(x1, y1)
                for info in curves[0].peak_info:
                    px = info[1]
                    ax1.plot(px, spline(px), "o", color="red")
                ax1.plot(A*x2+B, scale*y2, ":")
                for info in curves[1].peak_info:
                    px = info[1]
                    ux = A*px + B
                    ax1.plot(ux, scale*spline2(px), "o", color="cyan", markersize=20, alpha=0.3)

                def draw_slice(ax, slice_):
                    ymin, ymax = ax.get_ylim()
                    ax.set_ylim(ymin, ymax)
                    for j in [slice_.start, slice_.stop]:
                        ax.plot([j, j], [ymin, ymax], ":", color="yellow")

                draw_slice(ax1, curves[0].peak_slice)

                ax2.set_title("Improved Mapping")
                ax2.plot(x1, y1)
                for info in curves[0].peak_info:
                    px = info[1]
                    ax2.plot(px, spline(px), "o", color="red")
                ax2.plot(a*x2+b, scale*y2, ":")
                for info in curves[1].peak_info:
                    px = info[1]
                    ux = a*px + b
                    ax2.plot(ux, scale*spline2(px), "o", color="cyan", markersize=20, alpha=0.3)

                draw_slice(ax2, curves[0].peak_slice)

                fig.tight_layout()
                plt.show()

        return a, b

    def compare_primary_peak_mapping_params(self, curve1, curve2, mapping, debug=False):
        px1 = curve1.primary_peak_x
        px2 = curve2.primary_peak_x
        scale = curve1.spline(px1)/curve2.spline(px2)
        uv_x = curve1.x
        x2 = curve2.x

        def objective(p, add_no_mapping=False, debug=False):
            a, b, w = p
            if w <= 1e-3:
                return np.inf

            x1 = a*x2 + b
            compare_interval = np.logical_and(px1 - w <= x1, x1 <= px1 + w)
            x1_ = x1[compare_interval]
            x2_ = x2[compare_interval]
            dev = np.sum((curve1.spline(x1_) - scale*curve2.spline(x2_))**2)

            if add_no_mapping:
                # add penalty from region with no mapping as in OA_Ald with different primary peaks
                if x1[0] < uv_x[0]:
                    extend_L = np.arange(x1[0], uv_x[0], 1)   # uv_x[1] - uv_x[0] == 1
                else:
                    extend_L = []
                if uv_x[-1] < x1[-1]:
                    extend_R = np.arange(uv_x[-1]+1, x1[-1]+1, 1)
                else:
                    extend_R = []
                ex_uv_x = np.concatenate([extend_L, uv_x, extend_R])
                penalty = np.sum(curve1.spline(ex_uv_x[ex_uv_x < x1[0]])**2) + np.sum(curve1.spline(ex_uv_x[ex_uv_x > x1[-1]])**2)
            else:
                penalty = 0

            if debug:
                print("dev/w=", dev/w)
                print("penalty=", penalty)
                with plt.Dp():
                    fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
                    fig.suptitle("objective debug")

                    simple_plot(ax1, curve1, legend=False)
                    simple_plot(ax2, curve2, legend=False)

                    y1_ = curve1.spline(x1_)
                    y2_ = curve2.spline(x2_)
                    ax1.plot(x1_, y1_, ":", label="compare_interval")
                    ax2.plot(x2_, y2_, ":", label="compare_interval")
                    ax1.fill_between(x1_, y1_, scale*y2_, color='pink', alpha=0.2)

                    for ax in ax1, ax2:
                        ax.legend()

                    fig.tight_layout()
                    plt.show()

            return dev/w + penalty

        a_init = len(curve1.x)/len(curve2.x)
        b_init = px1 - a_init*px2
        pinfo = curve1.peak_info[curve1.primary_peak_no]
        w_init = (pinfo[2] - pinfo[0])/2
        ret = minimize(objective, (a_init, b_init, w_init))
        a, b, w = ret.x
        if debug:
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title("compare_primary_peak_mapping_params")
                simple_plot(ax, curve1)
                x1 = a*x2 + b
                ax.plot(x1, scale*curve2.y, ":")
                fig.tight_layout()
                plt.show()

        W = len(curve1.x)
        primary_score = objective((a, b, W), add_no_mapping=True)
        A, B = mapping
        compare_score = objective((A, B, W), add_no_mapping=True)
        if primary_score >= compare_score:
            self.logger.info("adopt normal peak mapping with score comparison %.3g <= %.3g", compare_score, primary_score)
            a, b = A, B
        else:
            self.logger.info("adopt primary peak mapping with score comparison %.3g < %.3g", primary_score, compare_score)

        return a, b



"""
    DraggableArrow.py

    ref:
        https://stackoverflow.com/questions/63561034/arrows-with-draggable-end-in-matplotlib

    Copyright (c) 2022, SAXS Team, KEK-PF
"""

from matplotlib import pyplot as plt
from matplotlib.backend_bases import MouseButton
from matplotlib.patches import ArrowStyle

class DraggableArrow:
    def __init__(self, event):
        ax = event.inaxes
        self.ax = ax
        x = event.xdata
        y = event.ydata
        end_x = x+0.5
        end_y = y
        color = "C0"
        arrowstyle = ArrowStyle('<|-', head_width=0.4, head_length=0.6)
        self.vector = ax.annotate("", xy=(x, y), xytext=(end_x, end_y),
                                    arrowprops=dict(arrowstyle=arrowstyle,
                                                    fc=color, ec=color,
                                                    shrinkA=0, shrinkB=0)
                                    )
        self.start_point, = ax.plot(x, y, "o", color=color)
        self.end_point, = ax.plot(end_x, end_y, ".", color='none', picker=True)

    def update_position(self):
        ex, ey = self.end_point.get_data()
        self.vector.set_position([ex[0], ey[0]])

arrow = None
picked_artist = None

def button_press_callback(event):
    global picked_artist, arrow
    print("button_press_callback")
    if arrow is None:
        arrow = DraggableArrow(event)
        event.inaxes.figure.canvas.draw()
        print("created an arrow")

def pick_callback(event):
    'called when an element is picked'
    global picked_artist
    print("pick_callback")
    if event.mouseevent.button == MouseButton.LEFT:
        picked_artist = event.artist

def button_release_callback(event):
    'called when a mouse button is released'
    global picked_artist
    print("button_release_callback")
    if event.button == MouseButton.LEFT:
        picked_artist = None

def motion_notify_callback(event):
    'called when the mouse moves'
    global picked_artist
    # print("motion_notify_callback")
    if picked_artist is not None and event.inaxes is not None and event.button == MouseButton.LEFT:
        print("motion_notify_callback: move")
        picked_artist.set_data([event.xdata], [event.ydata])
        arrow.update_position()
        event.inaxes.figure.canvas.draw_idle()

def demo():
    fig, ax = plt.subplots()

    ax.plot(0, 0, "o")

    ax.set_xlim(-1, 1)
    ax.set_ylim(-1, 1)

    fig.canvas.mpl_connect('button_press_event', button_press_callback)
    fig.canvas.mpl_connect('button_release_event', button_release_callback)
    fig.canvas.mpl_connect('pick_event', pick_callback)
    fig.canvas.mpl_connect('motion_notify_event', motion_notify_callback)

    plt.show()

if __name__ == '__main__':
    demo()



"""
    Theory.SynthesizedLRF.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import numpy as np
from bisect import bisect_right
from scipy.optimize import minimize
from matplotlib.patches import Rectangle
from time import time
from molass_legacy.KekLib.SciPyCookbook import smooth
import molass_legacy.KekLib.DebugPlot as plt
from SvdDenoise import get_denoised_data
from .SolidSphere import phi
from .Rg import compute_corrected_Rg
from .SolidSphere import get_boundary_params_simple

def synthesized_lrf_spike(qv, M, c, M2, P2, boundary=None, k=1):
    w = 1/(1+np.exp(-k*(qv-boundary)))

    M1 = get_denoised_data(M, rank=1)
    C1 = np.array([c])
    P1 = M1 @ np.linalg.pinv(C1)

    aq = P2[:,0] * (1-w) + P1[:,0] * w
    bq = P2[:,1] * (1-w) + np.zeros(P1.shape[0]) * w
    return np.array([aq, bq]).T

def demo(root, sd, pno=0, debug=False):
    from .BoundedLRF import demo as demo_impl
    demo_impl(root, sd, pno=pno, synthesized_lrf=synthesized_lrf_spike, bounded_lrf=True, debug=debug)

def get_reduced_conc(C_, cdl_list):
    c_list = []
    for k, cd in enumerate(cdl_list):
        if cd == 2:
            c_list.append(C_[k,:])
    return np.array(c_list)

def synthesized_lrf(qv, D, C, rank, Cred, Rg):
    t0 = time()

    D2 = get_denoised_data(D, rank=rank)
    P2 = D2 @ np.linalg.pinv(C)

    red_rank = len(Cred)
    D1 = get_denoised_data(D, rank=red_rank)
    P1 = D1 @ np.linalg.pinv(Cred)

    b1, b2, k = get_boundary_params_simple(Rg)
    w = 1/(1+np.exp(-k*(qv-b1)))

    aq = P2[:,0] * (1-w) + P1[:,0] * w
    bq = P2[:,1] * (1-w) + np.zeros(P1.shape[0]) * w

    # this is to be compatible with ExtrapolationSolver.fit
    # opt_info = [num_iterations, func_calls, time_elapsed]
    opt_info = [0, 0, time() - t0]

    P = np.array([aq, bq]).T
    return P, opt_info, D2, b1



"""

    Molecular dynamics in Python
    https://nznano.blogspot.com/2017/11/molecular-dynamics-in-python.html

"""
#Import a plotting libraries and a maths library 
import matplotlib.pyplot as plt
import numpy as np
# import molass_legacy.KekLib.DebugPlot as plt
# %matplotlib inline

r = np.linspace(0.01,3.0,num=500) # Make a radius vector
epsilon = 1 # Energy minimum
sigma = 1 # Distance to zero crossing point
E_LJ = 4*epsilon*((sigma/r)**12-(sigma/r)**6) # Lennard-Jones potential

plt.figure(figsize=[6,6])
plt.plot(r,E_LJ,'r-',linewidth=1,label=r" $LJ\; pot$") # Red line is unshifted LJ

# The cutoff and shifting value
Rcutoff = 2.5
phicutoff = 4.0/(Rcutoff**12)-4.0/(Rcutoff**6) # Shifts the potential so at the cutoff the potential goes to zero

E_LJ_shift = E_LJ - phicutoff # Subtract the value of the potential at r=2.5

if False:
    plt.plot(r[:415],E_LJ_shift[:415],'b-',linewidth=1,label=r"$LJ\; pot\; shifted$") # Blue line is shifted

    #Plot formatting
    # plt.rc('text', usetex=True)
    plt.rc('xtick', labelsize=20) 
    plt.rc('ytick', labelsize=20) 
    plt.title(r"$Lennard-Jones\; potential$",fontsize=20)
    plt.xlim([0.0,3.0])
    plt.ylim([-1.5,1.5])
    plt.ylabel(r"$E_{LJ}/\epsilon$",fontsize=20)
    plt.xlabel(r"$r/\sigma$",fontsize=20)
    plt.legend(frameon=False,fontsize=20)
    plt.axhline(0, color='grey',linestyle='--',linewidth=2)
    plt.axvline(1, color='grey',linestyle='--',linewidth=2)
    plt.show()

def Compute_Forces(pos,acc,ene_pot,epsilon,BoxSize,DIM,N):
    # Compute forces on positions using the Lennard-Jones potential
    # Uses double nested loop which is slow O(N^2) time unsuitable for large systems
    Sij = np.zeros(DIM) # Box scaled units
    Rij = np.zeros(DIM) # Real space units
    
    #Set all variables to zero
    ene_pot = ene_pot*0.0
    acc = acc*0.0
    virial=0.0
    
    # Loop over all pairs of particles
    for i in range(N-1):
        for j in range(i+1,N): #i+1 to N ensures we do not double count
            Sij = pos[i,:]-pos[j,:] # Distance in box scaled units
            for l in range(DIM): # Periodic interactions
                if (np.abs(Sij[l])>0.5):
                    Sij[l] = Sij[l] - np.copysign(1.0,Sij[l]) # If distance is greater than 0.5  (scaled units) then subtract 0.5 to find periodic interaction distance.
            
            Rij = BoxSize*Sij # Scale the box to the real units in this case reduced LJ units
            Rsqij = np.dot(Rij,Rij) # Calculate the square of the distance
            
            if(Rsqij < Rcutoff**2):
                # Calculate LJ potential inside cutoff
                # We calculate parts of the LJ potential at a time to improve the efficieny of the computation (most important for compiled code)
                rm2 = 1.0/Rsqij # 1/r^2
                rm6 = rm2**3.0 # 1/r^6
                rm12 = rm6**2.0 # 1/r^12
                phi = epsilon*(4.0*(rm12-rm6)-phicutoff) # 4[1/r^12 - 1/r^6] - phi(Rc) - we are using the shifted LJ potential
                # The following is dphi = -(1/r)(dV/dr)
                dphi = epsilon*24.0*rm2*(2.0*rm12-rm6) # 24[2/r^14 - 1/r^8]
                ene_pot[i] = ene_pot[i]+0.5*phi # Accumulate energy
                ene_pot[j] = ene_pot[j]+0.5*phi # Accumulate energy
                virial = virial + dphi*np.sqrt(Rsqij) # Virial is needed to calculate the pressure
                acc[i,:] = acc[i,:]+dphi*Sij # Accumulate forces
                acc[j,:] = acc[j,:]-dphi*Sij # (Fji=-Fij)
    return acc, np.sum(ene_pot)/N, -virial/DIM # return the acceleration vector, potential energy and virial coefficient

def Calculate_Temperature(vel,BoxSize,DIM,N):
    
    ene_kin = 0.0
    
    for i in range(N):
        real_vel = BoxSize*vel[i,:]
        ene_kin = ene_kin + 0.5*np.dot(real_vel,real_vel)
    
    ene_kin_aver = 1.0*ene_kin/N
    temperature = 2.0*ene_kin_aver/DIM
    
    return ene_kin_aver,temperature

DIM = 2 # Dimensions
N = 32

BoxSize = 10.0   #6.35 

volume  = BoxSize**DIM
density = N / volume
print("volume = ", volume, " density = ", density)

# pos = np.zeros([N,DIM])
        
# pos = np.genfromtxt('output.dat',skip_header=1) # Load positions from file
pos = np.random.uniform(-BoxSize*0.5, BoxSize*0.5, (N,DIM))
pos = pos[:,:DIM]/BoxSize

MassCentre = np.sum(pos,axis=0)/N

for i in range(DIM):
    pos[:,i] = pos[:,i]-MassCentre[i]

# Setting up the simulation
NSteps=10000 # Number of steps
deltat = 0.0032 # Time step in reduced time units
TRequested = 0.5# #Reduced temperature
DumpFreq = 100 # Save the position to file every DumpFreq steps
epsilon = 1.0 # LJ parameter for the energy between particles

# Main MD loop
def main(pos,NSteps,deltat,TRequested,DumpFreq,epsilon,BoxSize,DIM,notebook=False,save_traj=False):
    if not notebook:
        pos_list = []

    # Vectors to store parameter values at each step
    N = np.size(pos[:,1])
    ene_kin_aver = np.ones(NSteps)
    ene_pot_aver = np.ones(NSteps)
    temperature = np.ones(NSteps)
    virial = np.ones(NSteps)
    pressure = np.ones(NSteps)
    ene_pot = np.ones(N)

    vel = (np.random.randn(N,DIM)-0.5)
    acc = (np.random.randn(N,DIM)-0.5)

    if save_traj:
        # Open file which we will save the outputs to
        f = open('traj.xyz', 'w')
    
    for k in range(NSteps):
        
        # Refold positions according to periodic boundary conditions
        for i in range(DIM):
            period = np.where(pos[:,i] > 0.5)
            pos[period,i]=pos[period,i]-1.0
            period = np.where(pos[:,i] < -0.5)
            pos[period,i]=pos[period,i]+1.0

        # r(t+dt) modify positions according to velocity and acceleration
        pos = pos + deltat*vel + 0.5*(deltat**2.0)*acc # Step 1

        # Calculate temperature
        ene_kin_aver[k],temperature[k] = Calculate_Temperature(vel,BoxSize,DIM,N)

        # Rescale velocities and take half step
        chi = np.sqrt(TRequested/temperature[k])
        vel = chi*vel + 0.5*deltat*acc # v(t+dt/2) Step 2

        # Compute forces a(t+dt),ene_pot,virial
        acc, ene_pot_aver[k], virial[k] = Compute_Forces(pos,acc,ene_pot,epsilon,BoxSize,DIM,N) # Step 3

        # Complete the velocity step 
        vel = vel + 0.5*deltat*acc # v(t+dt/2) Step 4

        # Calculate temperature
        ene_kin_aver[k],temperature[k] = Calculate_Temperature(vel,BoxSize,DIM,N)

        # Calculate pressure
        pressure[k]= density*temperature[k] + virial[k]/volume
        
        
        # Print output to file every DumpFreq number of steps
        if(k%DumpFreq==0): # The % symbol is the modulus so if the Step is a whole multiple of DumpFreq then print the values

            if save_traj:
                f.write("%s\n" %(N)) # Write the number of particles to file
                # Write all of the quantities at this step to the file
                f.write("Energy %s, Temperature %.5f\n" %(ene_kin_aver[k]+ene_pot_aver[k],temperature[k]))
                for n in range(N): # Write the positions to file
                    f.write("X"+" ")
                    for l in range(DIM):
                        f.write(str(pos[n][l]*BoxSize)+" ")
                    f.write("\n")
            
            if(DIM==2):
                if notebook:
                    import matplotlib.pyplot as plt
                    from IPython import display
                    plt.cla()
                    plt.xlim(-0.5*BoxSize,0.5*BoxSize)
                    plt.ylim(-0.5*BoxSize,0.5*BoxSize)
                    for i in range(N):
                        plt.plot(pos[i,0]*BoxSize,pos[i,1]*BoxSize,'o',markersize=20,)
                    display.clear_output(wait=True)
                    display.display(plt.gcf())
                else:
                    print([k])
                    pos_list.append(pos.copy())
        #print(ene_kin_aver[k], ene_pot_aver[k], temperature[k], pressure[k]) 

    if save_traj:
        f.close() # Close the file

    if notebook:
        plt.close()
    else:
        aminate(pos_list, BoxSize)
    return ene_kin_aver, ene_pot_aver, temperature, pressure

def aminate(pos_list, BoxSize):
    from matplotlib.animation import FuncAnimation

    # fig, ax = plt.subplots()
    fig = plt.figure()
    ax = fig.gca()
    ax.set_xlim(-0.5*BoxSize,0.5*BoxSize)
    ax.set_ylim(-0.5*BoxSize,0.5*BoxSize)

    pos = pos_list[0]
    arts, = ax.plot(pos[:,0]*BoxSize, pos[:,1]*BoxSize, 'o', markersize=20)

    def init():
        return arts,

    def update(i):
        pos = pos_list[i]
        arts.set_data(pos[:,0]*BoxSize, pos[:,1]*BoxSize)
        return arts,

    fig.tight_layout()

    ani = FuncAnimation(fig, update, frames=len(pos_list),
                        init_func=init, blit=True)

    plt.show()

def demo():
    ene_kin_aver, ene_pot_aver, temperature, pressure = main(pos,NSteps,deltat,TRequested,DumpFreq,epsilon,BoxSize,DIM)

# ------------------------------------------------------------------------------

def simulate(DIM=2, N=32, BoxSize=10.0):

    volume  = BoxSize**DIM
    density = N / volume
    print("volume = ", volume, " density = ", density)

    # pos = np.zeros([N,DIM])
            
    # pos = np.genfromtxt('output.dat',skip_header=1) # Load positions from file
    pos = np.random.uniform(-BoxSize*0.5, BoxSize*0.5, (N,DIM))
    pos = pos[:,:DIM]/BoxSize

    MassCentre = np.sum(pos,axis=0)/N

    for i in range(DIM):
        pos[:,i] = pos[:,i]-MassCentre[i]

    # Setting up the simulation
    NSteps = 10000 # Number of steps
    deltat = 0.0032 # Time step in reduced time units
    TRequested = 0.5# #Reduced temperature
    DumpFreq = 100 # Save the position to file every DumpFreq steps
    epsilon = 1.0 # LJ parameter for the energy between particles

    ene_kin_aver, ene_pot_aver, temperature, pressure = main(pos, NSteps, deltat, TRequested, DumpFreq, epsilon, BoxSize, DIM)



"""
    Wave.Beam.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import matplotlib.animation as animation
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar

class BeamAnimDialog(Dialog):
    def __init__(self, parent):
        Dialog.__init__(self, parent, "Beam Animatio", visible=False)

    def show(self):
        self._show()

    def body(self, body_frame):
        cframe = Tk.Frame(body_frame)
        cframe.pack()
        bframe = Tk.Frame(body_frame)
        bframe.pack(fill=Tk.BOTH, expand=1)
        tframe = Tk.Frame(bframe)
        tframe.pack(side=Tk.LEFT)
        pframe = Tk.Frame(bframe)
        pframe.pack(side=Tk.RIGHT)

        fig, axes = plt.subplots(ncols=2, figsize=(14, 7))
        self.fig  = fig
        self.mpl_canvas = FigureCanvasTkAgg(fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)
        self.toolbar = NavigationToolbar(self.mpl_canvas, tframe)
        self.toolbar.update()
        self.draw_anim(fig, axes)
        self.mpl_canvas.draw()

    def draw_anim(self, fig, axes):
        ax1, ax2 = axes

        x = np.linspace(0, 10*np.pi, 100)
        y = np.sin(x)

        ax1.set_ylim(-3, 3)
        line, = ax1.plot(x, y)

        ax2.set_ylim(0,1)
        z = np.ones(len(x))*0.5
        artists = ax2.scatter(x, z, vmin=-1, vmax=1, c=y, cmap=cm.plasma)

        fig.tight_layout()

        def init():
          line.set_data([],[])
          artists.set_array(y)
          return [line, artists]

        num_frames = 40

        def animate(i):
          y = np.sin(x-10*np.pi*i/num_frames)
          line.set_data(x,y)
          artists.set_array(y)
          return [line, artists]

        self.anim = animation.FuncAnimation(fig, animate, init_func=init,
                                        frames=num_frames, interval=100, blit=True)

        # self.anim.save('beam-wave.mp4', writer="ffmpeg")

def demo1():
    from molass_legacy.KekLib.TkUtils import get_tk_root

    root = get_tk_root()

    def show_demo():
        dialog = BeamAnimDialog(root)
        dialog.show()
        root.quit()

    root.after(100, show_demo)
    root.mainloop()
    root.destroy()



"""
    Distance.DeviationMeasure.py

    Copyright (c) 2023-2024, SAXS Team, KEK-PF
"""
import numpy as np
import molass_legacy.KekLib.DebugPlot as plt

VALID_Y_RATIO = 0.03
WANTED_RATIO_LIMIT = 0.5
FEATURE_SCALE = 2

def feature_deviation(y1, y2, max_y2, y2_norm, debug=False, fig_info=None):
    diff = y1 - y2
    base = 0.7*y1 + 0.3*y2
    valid = y2 > max_y2*VALID_Y_RATIO
    ratio = np.abs(diff[valid])/base[valid]
    wanted_ratio = ratio[ratio > WANTED_RATIO_LIMIT]
    if len(wanted_ratio) > 0:
        fdev = np.mean(wanted_ratio)
    else:
        fdev = WANTED_RATIO_LIMIT
    ret_dev = np.log10(np.linalg.norm(diff)/y2_norm)/FEATURE_SCALE + np.log10(fdev)*FEATURE_SCALE

    if debug:
        print("fdev=", fdev)
        print("ret_dev=", ret_dev)
        x = np.arange(len(y2))
        x_ = x[valid]
        def debug_plot(fig, ax):
            ax.set_title("feature_deviation debug")
            ax.plot(x, y1, label="y1")
            ax.plot(x, y2, label="y2")
            ax.plot(x, base, label="base")
            ax.plot(x, diff, label="diff")
            axt = ax.twinx()
            axt.grid(False)
            axt.plot(x_, ratio, label="ratio")
            ax.legend()
            axt.legend(loc="center right")

        if fig_info is None:
            with plt.Dp():
                fig, ax = plt.subplots()
                debug_plot(fig, ax)
                fig.tight_layout()
                plt.show()
        else:
            fig, ax = fig_info
            debug_plot(fig, ax)
    
    return ret_dev



"""
    Theory.PbMoore1980.py

    Basic functions shannon_channels, Bt, Ct, Yt, ...
    are borrowed from saxstats.py in DENSS

        Author: Thomas D. Grant
        Email:  <tdgrant@buffalo.edu>
        Alt Email:  <tgrant@hwi.buffalo.edu>
        Copyright 2017, 2018, 2019, 2020 The Research Foundation for SUNY
"""
import numpy as np
import molass_legacy.KekLib.DebugPlot as plt

"""
    Y = 4*A@C
    A = 1/4*Y@C⁻¹
"""

def shannon_channels(D, qmax=0.5, qmin=0.0):
    """Return the number of Shannon channels given a q range and maximum particle dimension"""
    width = np.pi / D
    num_channels = int((qmax-qmin) / width)
    return num_channels

def Bt(n, q, D):
    return (n*np.pi)**2/((n*np.pi)**2-(q*D)**2) * np.sinc(q*D/np.pi) * (-1)**(n+1)
    # return np.pi*n*D*(-1)**(n+1)*np.sin(2*np.pi*D*q)/((np.pi*n)**2 - (2*np.pi*D*q)**2)

def Yt(I, Ierr, Bm):
    """Return the values of Y, an m-length vector."""
    return np.einsum('i, ki->k', I/Ierr**2, Bm)

def Ct(Ierr, Bm, Bn):
    """Return the values of C, a m x n variance-covariance matrix"""
    return np.einsum('ij,kj->ik', Bm/Ierr**2, Bn)

def shannon_intensities(q, I, Ierr, D, ne=2):
    """Calculate Shannon intensities from experimental I(q) profile."""

    nq = len(q)
    qmin = q[0]
    qmax = q[-1]
    n = shannon_channels(qmax, D) + ne

    B = np.zeros((n, nq))
    C = np.zeros((n, n))
    Y = np.zeros((n))

    Ni = np.arange(n)
    N = Ni + 1
    Mi = Ni.copy()
    M = N.copy()
    qn = np.pi/D * N

    B[Ni] = Bt(N[:, None], q, D)
    Y[Mi] = Yt(I, Ierr, B[Mi])
    C[Mi[:, None], Ni] = Ct(Ierr, B[Mi], B[Ni])
    Cinv = np.linalg.inv(C)
    In = 0.5 * np.linalg.solve(C, Y)
    Inerr = 0.5 * (np.diagonal(Cinv))**(0.5)
    return qn, In, Inerr

def Ish2Iq(Ish, D, q=(np.arange(500)+1.)/1000):
    """Calculate I(q) from intensities at Shannon points."""
    q = np.atleast_1d(q)
    Ish = np.atleast_1d(Ish)
    Iq = np.zeros((len(q),2))
    Iq[:,0] = q
    n = len(Ish)
    N = np.arange(n)+1
    denominator = (N[:,None]*np.pi)**2-(q*D)**2
    I = 2*np.einsum('k,ki->i',Ish,(N[:,None]*np.pi)**2 / denominator * np.sinc(q*D/np.pi) * (-1)**(N[:,None]+1))
    Iq[:,1] = I
    return Iq

def ift_intensities(q, I, Ierr, D=None, qc=None, r=None, ne=2):
    if D is None:
        from molass_legacy.DENSS.saxstats.saxstats import estimate_dmax
        Iq = np.array([q, I, Ierr]).T
        D = estimate_dmax(Iq)[0]
        print("estimated dmax = %.4g" % D)

    qn, In, Inerr = shannon_intensities(q, I, Ierr, D, ne=ne)

    if qc is None:
        qc = q
    Ic = Ish2Iq(In, D, q=qc)[:,1]
    Icerr = np.interp(qc, q, Ierr)
    return qc, Ic, Icerr, D

"""
    Copyright (c) 2021-2023, SAXS Team, KEK-PF
"""
def demo(root, sd, pno=0, debug=True, crysol_int_files=None):
    from bisect import bisect_right
    from .SynthesizedLRF import synthesized_lrf_spike as synthesized_lrf
    from SvdDenoise import get_denoised_data
    from .Rg import compute_corrected_Rg, compute_Rg
    from .SolidSphere import get_boundary_params_simple

    assert crysol_int_files is not None

    M, E, qv, ecurve = sd.get_xr_data_separate_ly()

    range_ = ecurve.get_ranges_by_ratio(0.5)[pno]
    f = range_[0]
    p = range_[1]
    t = range_[2]

    eslice = slice(f,t)
    x = ecurve.x
    y = ecurve.y

    M0 = M[:,eslice]

    c_ = y[eslice]
    M2 = get_denoised_data(M0, rank=2)
    C2 = np.array([c_, c_**2])
    P2 = M2 @ np.linalg.pinv(C2)

    Rg, gf, gt, _ = compute_corrected_Rg(sd, ecurve, pno, qv, M0, E[:,eslice])
    b1, b2, k = get_boundary_params_simple(Rg)
    P = synthesized_lrf(qv, M0, c_, M2, P2, boundary=b1, k=k)

    pt = M[:,p]
    spt = P[:,0]

    qc2, Ic2, Icerr2, D = ift_intensities(qv, spt, E[:,p])

    i = bisect_right(qv, b1)
    pt_ = pt*Ic2[i]/pt[i]
    ptc = np.hstack([Ic2[0:i], pt_[i:]])
    qc1, Ic1, Icerr1, _ = ift_intensities(qv, ptc, E[:,p], D=D)

    ncols = len(crysol_int_files) + 1

    fig, axes = plt.subplots(ncols=ncols, figsize=(7*ncols,7))
    fig.suptitle("Difference between ALMERGE and SynLRF with IFT from P. B. Moore 1980", fontsize=20)

    ax1, ax2 = axes[0:2]
    ax1.set_title("Peak Top, LRF and PBMoore", fontsize=16)
    if len(axes) == 2:
        ax2.set_title("PBMoore and CRYSOL", fontsize=16)
        ax3 = None
    else:
        ax3 = axes[2]
        ax2.set_title("PBMoore and CRYSOL2", fontsize=16)
        ax3.set_title("PBMoore and CRYSOL3", fontsize=16)

    for ax in axes:
        ax.set_yscale('log')

    ax1.plot(qv, pt_, label='Peak Top Data')
    ax1.plot(qv, spt, label='Syn LRF')

    ymin, ymax = ax1.get_ylim()

    for ax in axes:
        ax.set_ylim(ymin, ymax)
        ax.plot(qc1, Ic1, label="ALMERGE like IFT", color='C2')
        ax.plot(qc2, Ic2, label="Syn LRF IFT", color='C3')

    if crysol_int_files is not None:
        from CrysolUtils import np_loadtxt_crysol
        for j, ax in enumerate(axes[1:]):
            crysol_int = crysol_int_files[j]
            name = crysol_int[0:4]
            data, _ = np_loadtxt_crysol(crysol_int)
            print('data.shape=', data.shape)
            if data.shape[1] == 5:
                crysol_ver = 2
                K = 2
            else:
                crysol_ver = 3
                K = 2
            qv = data[:,0]
            for k in range(1,K):
                y = data[:,k].copy()
                y *= Ic2[0]/y[0]
                if ax3 is None:
                    ax.plot(qv, y, label='%s-CRYSOL' % (name), color='C%d' % (3+k))
                else:
                    # older style when there were tow versions of crysol, i.e., crysol and crysol_30
                    ax.plot(qv, y, label='%s-CRYSOL%d-%d' % (name, crysol_ver, k), color='C%d' % (3+k))

    for ax in axes:
        ax.plot([b1, b1], [ymin, ymax], ':', color='gray', label='Rank boundary')
        ax.legend()

    fig.tight_layout()
    fig.subplots_adjust(top=0.88)
    plt.show()



"""
    SecSaxs.ElCurve.py

    1) temporary successor to ElutionCurve
    2) x starts with a value at the slice start when trimmed

    Copyright (c) 2022-2025, SAXS Team, KEK-PF
"""
import logging
import os
import numpy as np
from scipy.interpolate import UnivariateSpline
import molass_legacy.KekLib.DebugPlot as plt

SEDIMENTATION_LIMIT = 0.03
END_DIFF_WIDTH = 50
END_DIFF_MIN_HEIGHT = SEDIMENTATION_LIMIT/2
RELIABLE_SIGMA_RATIO = 2
RELIABLE_ALPHA = 0.03
USE_RELIABLE_ALPHA = True

class ElCurve:
    def __init__(self, x, y, v1_curve=None, debug=False):
        self.logger = logging.getLogger(__name__)
        self.x = x
        self.y = y
        if v1_curve is None:
            from molass_legacy.SerialAnalyzer.ElutionCurve import ElutionCurve
            v1_curve = ElutionCurve(y, x=x)
        self.peak_top_x = v1_curve.peak_top_x + x[0]
        self.peak_info = [[x[0] + j for j in info] for info in v1_curve.peak_info]
        self.boundaries = [x[0] + j for j in v1_curve.boundaries]
        self.emg_peaks = [epeak.shift_copy(x[0], x[-1]+1) for epeak in v1_curve.emg_peaks]       
        self.sy = v1_curve.sy
        print("len(x), len(y), len(sy):", len(x), len(y), len(self.sy))
        self.spline = UnivariateSpline(x, self.sy, s=0, ext=3)
        self.primary_peak_no = v1_curve.primary_peak_no
        if v1_curve is None or True:
            j = np.argmax(y)
            self.max_j = j
            self.max_x = x[j]
            self.max_y = y[j]
        else:
            self.max_x = v1_curve.x[v1_curve.max_x]
            self.max_y = v1_curve.max_y
        self.mean = None
        self.variance = None
        self.sigma = None
        self.end_slices = None
        self.primary_peak_x = self.max_x    # temporary fix for 20250212/SG
        self.peak_region_sigma = None       # temporary fix for 20250212/SG

        if debug:
            with plt.Dp():
                fig, axes = plt.subplots(ncols=2, figsize=(12,5))
                fig.suptitle("ElCurve.__init__")
                for ax, curve in zip(axes, [v1_curve, self]):
                    x = curve.x
                    y = curve.y
                    ax.plot(x, y)
                    for px in curve.peak_top_x:
                        ax.plot(px, curve.spline(px), "o", color="red")
                fig.tight_layout()
                plt.show()

    def get_xy(self):
        # for forward compatibility
        return self.x, self.y

    def get_primarypeak_x(self):
        return self.max_x

    def get_primarypeak_i(self):
        return int(round(self.max_x - self.x[0]))
    
    def compute_moments(self):
        if self.mean is None:
            x = self.x
            y = np.zeros(len(x))

            # self.y can be negative as in 20200123_1
            positive_y = self.y > 0
            y[positive_y] = self.y[positive_y]

            sum_y = np.sum(y)
            mean  = np.sum(x*y)/sum_y
            variance = np.sum(x**2*y)/sum_y - mean**2
            self.mean = mean
            self.variance = variance
            self.sigma = np.sqrt(variance)

        return self.mean, self.variance

    def get_sigma(self):
        if self.sigma is None:
            self.compute_moments()

        return self.sigma

    def get_emg_peaks(self, **kwargs):
        return self.emg_peaks

    def get_peak_region_sigma(self):
        if self.peak_region_sigma is None:
            emg_peaks = self.get_emg_peaks()
            h1, m1, s1, t1 = emg_peaks[0].get_params()
            h2, m2, s2, t2 = emg_peaks[-1].get_params()
            peak_region = (max(0, m1-3*s1), min((len(self.x)), m2+3*s2))
            self.peak_region_sigma = peak_region
        return self.peak_region_sigma

    def get_end_y(self, slice_):
        y_ = self.y[slice_]
        h = abs(y_[-1] - y_[0])
        ok = False
        if h/self.max_y < END_DIFF_MIN_HEIGHT:
            ok = True
        else:
            std = np.std(y_)
            if std/self.max_y < END_DIFF_MIN_HEIGHT:
                ok = True
        if ok:
            self.end_slices.append(slice_)
            return np.average(y_)

        # in this case, slice_ must be narrowed as obsesrved in 20181204
        if slice_.start == 0:
            return self.get_end_y(slice(0, slice_.stop//2))
        else:
            return self.get_end_y(slice(slice_.start//2, None))

    def compute_sedimentation_rate(self, debug=False):
        emg_peaks = self.get_emg_peaks()
        self.end_slices = []
        start_y = self.get_end_y(slice(0, END_DIFF_WIDTH))
        end_y = self.get_end_y(slice(-END_DIFF_WIDTH, None))
        sedrate = (end_y - start_y)/(self.max_y - start_y)

        slice1 = self.end_slices[0]
        width1 = slice1.stop

        slice2 = self.end_slices[1]
        width2 = abs(slice2.start)
        start2 = len(self.x) + slice2.start

        num_emg_peaks = len(emg_peaks)

        if num_emg_peaks > 0:
            # to reduce the calulation
            indeces = set([0, num_emg_peaks-1])

            lim_x_pairs = []
            for i in indeces:
                if USE_RELIABLE_ALPHA:
                    pair = emg_peaks[i].get_model_x_from_ratio(RELIABLE_ALPHA)
                else:
                    pair = emg_peaks[i].get_sigma_points(RELIABLE_SIGMA_RATIO)
                lim_x_pairs.append(pair)
            sigma_x1 = lim_x_pairs[0][0]
            sigma_x2 = lim_x_pairs[-1][1]
            reliable = (width1 >= END_DIFF_WIDTH//8 and slice1.stop < sigma_x1) and (width2 >= END_DIFF_WIDTH//8 and sigma_x2 < start2)
            if debug:
                print(self.end_slices, sigma_x1, sigma_x2, start2)
                print(width1, width2)

                x = self.x
                y = self.y

                plt.push()
                fig, ax = plt.subplots()
                ax.plot(x, y, label='data')
                for slice_ in [slice(0,END_DIFF_WIDTH), slice(-END_DIFF_WIDTH, None)]:
                    ax.plot(x[slice_], y[slice_], 'o', color='yellow')

                for peak in emg_peaks:
                    print(self.j0, peak, peak.get_sigma_points(RELIABLE_SIGMA_RATIO))

                for k in set([0, len(emg_peaks)-1]):
                    peak = emg_peaks[k]
                    ax.plot(x, peak.get_model_y(x), label='model')

                ymin, ymax = ax.get_ylim()
                ax.set_ylim(ymin, ymax)
                for sx in [sigma_x1, sigma_x2]:
                    ax.plot([sx, sx], [ymin, ymax], ':', color='gray')

                ax.legend()
                fig.tight_layout()
                plt.show()
                plt.pop()
        else:
            # temporary fix for pH6 to avoid a bug with the revised EmgPeak
            reliable = False

        self.logger.info("computed sedimentation rate as %.3g which is%s reliable", sedrate, "" if reliable else " not")
        return sedrate, reliable

    def get_end_slices(self):
        if self.end_slices is None:
            self.compute_sedimentation_rate()
        return self.end_slices



# coding: utf-8
"""
    SecTheory.Dondi2002.py

    Copyright (c) 2022, SAXS Team, KEK-PF
"""
import numpy as np

def monopore_gec(w, np_, tp_):
    return np.exp(np_*(1/(1 - 1j*w*tp_) - 1))

def twopore_gec(w, npa, tpa, npb, tpb):
    return np.exp(npa*(1/(1 - 1j*w*tpa) - 1) + npb*(1/(1 - 1j*w*tpb) - 1))

def dispersive_sec(w, np_, tp_, N0, t0):
    Z = np_*(1/(1 - 1j*w*tp_) - 1) + 1j*w*t0
    return np.exp(Z + 1/(2*N0)*Z**2)



"""
    InvertibleSpline.py

    Copyright (c) 2023-2024, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.interpolate import UnivariateSpline

class InvertibleSpline(UnivariateSpline):
    def __init__(self, x, y, k=3, s=None, ext=0, debug=True):
        try:
            UnivariateSpline.__init__(self, x, y, k=k, s=s, ext=ext)
        except Exception as exc:
            # better to aviod this situation. namely, remove the cause which brought you here.

            import logging
            import molass_legacy.KekLib.DebugPlot as plt
            from molass_legacy.KekLib.ExceptionTracebacker import log_exception

            logger = logging.getLogger(__name__)
            log_exception(logger, "InvertibleSpline: ", n=10)

            if len(x) in [2,3]:
                # as in 20161006/OA01/Backsub with cum_components=5
                k = len(x) - 1
                UnivariateSpline.__init__(self, x, y, k=k, s=s, ext=ext)
            else:
                if len(x) == 1:
                    raise ValueError("len(x) == 1")

                diff = x[1:] - x[:-1]
                illegal = np.where(diff <= 0)[0]
                print("x=", x)
                print("y=", y)
                print("illegal=", illegal)

                if debug:
                    with plt.Dp():
                        fig, ax = plt.subplots()
                        ax.set_title("InvertibleSpline debug plot")
                        ax.plot(x, y)
                        ax.plot(x[illegal], y[illegal], "o", color="red")
                        fig.tight_layout()
                        plt.show()
                raise exc



"""
    UV.WYF_RatiosPlot.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import numpy as np
import matplotlib.pyplot as plt
import seaborn
seaborn.set()

SIGN_DICT = {
    "D" : -1,
    "C" :  0,
    "U" : +1,
    "L" : -1,
    "R" : +1,
    }

annotate_points = [
    ("$OA$",  0.981, 0.551, "RD"),
    ("$ALD$", 0.960, 0.433, "RD"),
    ("$ALD_{bump}$", 0.965,0.511, "RD"),
    ("$GI$",  0.965, 0.566, "LD"),
    ("$BSA$", 1.01,  0.611, "LU"),
    ("$HasA$",1.00,  0.715, "RU"),
    ("$HasA$",1.00,  0.715, "RU"),
    ("$PKS$", 0.977, 0.592, "LC"),
    ("$HypC$", 1.16, 0.605, "RC"),
    ("$C1015s$", 1.03, 1.09, "RU")
    ]

def plot_wyf_ratios(wyf_ratios_file):
    value_list = []
    with open(wyf_ratios_file) as fh:
        for line in fh:
            row = line[:-1].split(",")
            value_list.append([float(s) for s in row[1:]])
    R = np.array(value_list)

    fig, ax = plt.subplots(figsize=(8,8))
    ax.set_title("WYF Absorbance Ratios: $A_{258}/A_{280}$ vs. $A_{275}/A_{280}$", fontsize=16)

    ax.set_xlabel("$A_{275}/A_{280}$")
    ax.set_ylabel("$A_{258}/A_{280}$")

    ax.plot(R[:,0], R[:,1], "o", markersize=3)

    ax.set_xlim(0.1, 1.7)
    ax.set_ylim(0.1, 1.7)
    xmin, xmax = ax.get_xlim()
    ymin, ymax = ax.get_ylim()
    dx = (xmax - xmin)*0.15
    dy = (ymax - ymin)*0.15

    def get_text_xy(x, y, loc):
        return x + SIGN_DICT[loc[0]]*dx, y + SIGN_DICT[loc[1]]*dy

    for name, r1, r2, loc in annotate_points:
        ax.plot(r1, r2, "o", color="red", markersize=3)
        tx, ty = get_text_xy(r1, r2, loc)
        color = "orange" if name.find("bump") >= 0 else None
        arrow_color = color if color == "orange" else "k"
        ax.annotate(name, xy=(r1, r2), xytext=(tx, ty), color=color, ha='center', arrowprops=dict(arrowstyle="->", color=arrow_color))

    fig.tight_layout()
    plt.show()

if __name__ == '__main__':
    wyf_ratios_file = r"D:\TODO\20231218\wyf-ratios\wyf-ratios-save.csv"
    plot_wyf_ratios(wyf_ratios_file)



"""
    Processes.Dummy.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
from time import sleep

def dummy(si, pn):
    print("dummy", si)

    si.in_folder.set("dummy_folder")

    while True:
        sleep(1)
        si.value.value += 1
        si.set_procstate(pn, si.value.value)
        print("dummy", si)
